/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "revision.h"
#include "registry.h"
#include "option.h"
#include "context.h"
#include "registry.h"
#include "colour.h"
#include "math.h"
#include "registry.h"


////// The methods for the Context class //////

// The constructor
Context::Context(CWnd *NewWindow, Transform NewTransform)
{
  theWindow = NewWindow;

  dc = new CClientDC(theWindow);

  deleteDC = TRUE;

  theTransform = NewTransform;

  // Initatalise the selectable objects
  nextNewPen   = 0;
  nextNewBrush = 0;

  theDefaultPen   = NULL;
  theDefaultBrush = NULL;
  theDefaultFont  = NULL;

  int lp;
  for (lp=0;lp!=MAX_PENS;lp++)
	thePen[lp] = NULL;
  for (lp=0;lp!=MAX_BRUSHES;lp++)
	theBrush[lp] = NULL;

  allBlack = FALSE;
  allGrey  = FALSE;
}

Context::Context(CDC *NewDC, Transform NewTransform, CWnd *pWnd)
{
  theWindow = pWnd;

  dc = NewDC;

  deleteDC = FALSE;

  theTransform = NewTransform;

  // Initatalise the selectable objects
  nextNewPen   = 0;
  nextNewBrush = 0;

  theDefaultPen   = NULL;
  theDefaultBrush = NULL;
  theDefaultFont  = NULL;

  int lp;
  for (lp=0;lp!=MAX_PENS;lp++)
	thePen[lp] = NULL;
  for (lp=0;lp!=MAX_BRUSHES;lp++)
	theBrush[lp] = NULL;

  allBlack = FALSE;
  allGrey  = FALSE;
}


// The destructor
Context::~Context()
{
  int lp;

  // Unselect the gdi objects before they can be deleted
  SelectDefault();

  // Now delete the gdi objects
  font_map::iterator it = m_fonts.begin();
  while (it != m_fonts.end())
  {
	  delete (*it).second;
	  it ++;
  }
  for (lp=0;lp!=MAX_PENS;lp++)
	if (thePen[lp] != NULL) {
		delete thePen[lp];
	}
  for (lp=0;lp!=MAX_BRUSHES;lp++)
  	if (theBrush[lp] != NULL) {
		delete theBrush[lp];
	}

  // Delete the DC (if necessary)
  if (deleteDC)
	delete dc;
}

void Context::SelectDefault()
{
  // Unselect the gdi objects
  if (theDefaultPen != NULL) 
  	dc->SelectObject(theDefaultPen);
  if (theDefaultFont != NULL)
  	dc->SelectObject(theDefaultFont);
  if (theDefaultBrush != NULL)
  	dc->SelectObject(theDefaultBrush);
}


////// The selection of drawing objects //////

BOOL Context::SelectPen(LineStyle *pStyle, paint_options options)
{
	switch (options)
	{
	case draw_selected:
		return SelectPen(pStyle->Style,pStyle->Thickness,cSELECT );
	case draw_selectable:
		{
			COLORREF col = pStyle->Colour;
			int red = min(255, GetRValue( col ) + 128 );
			int green = min(255, GetGValue( col ) + 128 );
			int blue =  min(255, GetBValue( col ) + 128 );
			return SelectPen(pStyle->Style,pStyle->Thickness, RGB(red,green,blue) );
		}
	default:
		return SelectPen(pStyle->Style,pStyle->Thickness,pStyle->Colour);
	}	
}



BOOL Context::SelectPen(int Style,int Width,LONG Colour)
{
  int Selected = -1;

  if (allBlack)
	Colour = RGB(0,0,0);

  if (allGrey)
  	Colour = RGB( GetRValue(Colour) >> 1, GetGValue(Colour) >> 1, GetBValue(Colour) >> 1);

  
  if (Width > 1 || dc->IsPrinting())
  {
	Width = theTransform.Scale(Width);
  }
  
  // Does this pen already exist?
  for (int lp =0;thePen[lp] != NULL && lp<MAX_PENS; lp++)
  {
	if (thePenColour[lp] == Colour && thePenWidth[lp] == Width && thePenStyle[lp] == Style) {
		Selected = lp;
		break;	
	}
  }

	// If it doesn't then create it
	if (Selected == -1) 
	{
		if (thePen[nextNewPen] != NULL) 
		{
			if (theDefaultPen != NULL)
	  			dc->SelectObject(theDefaultPen);
			delete thePen[nextNewPen];
		}



		thePen[nextNewPen] = new CPen(); 

		int new_style = Width <= 1 ? Style : Style | PS_GEOMETRIC;

		LOGBRUSH lb;
		lb.lbColor = Colour;
		lb.lbHatch = 0;
		lb.lbStyle = BS_SOLID;
		thePen[nextNewPen]->Attach( ::ExtCreatePen( new_style, Width, &lb, 0, NULL ) );
		thePenStyle[nextNewPen] = Style;
		thePenWidth[nextNewPen] = Width;
		thePenColour[nextNewPen] = Colour;
		Selected = nextNewPen;
		nextNewPen++;
		if (nextNewPen == MAX_PENS)
			nextNewPen = 0;
	}


  // Now select the pen
  CPen *r = dc->SelectObject(thePen[Selected]);
  if (theDefaultPen == NULL)
  {
	theDefaultPen = r;
  }


  return TRUE;
}




BOOL Context::SelectBrush(LONG Colour, int Index)
{
  int Selected = -1;

  if (allBlack)
	Colour = RGB(0,0,0);

  // Does this brush already exist?
  for (int lp = 0;lp<MAX_BRUSHES && theBrush[lp]!=NULL;lp++) 
  {
	if (   theBrushColour[lp] == Colour 
		&& theBrushIndex[lp] == Index 
		&& theBrush[lp] != NULL) {
		Selected = lp;
		break;
	}
  }

  // If it doesn't then create it
  if (Selected == -1) 
  {
	if (theBrush[nextNewBrush] != NULL) 
	{
		if (theDefaultBrush != NULL)
		{
	  		dc->SelectObject(theDefaultBrush);
		}
		delete theBrush[nextNewBrush];   
	}

	if (Index == 0)
	{
		theBrush[nextNewBrush] = new CBrush(Colour);
	}
	else
	{
		theBrush[nextNewBrush] = new CBrush(Index-1,Colour);
	}
	theBrushColour[nextNewBrush] = Colour;
	theBrushIndex[nextNewBrush] = Index;
	Selected = nextNewBrush;
	nextNewBrush++;
	if (nextNewBrush == MAX_BRUSHES)
		nextNewBrush = 0;
  }

  // Now select the brush
  CBrush *r = dc->SelectObject(theBrush[Selected]);
  if (theDefaultBrush == NULL)
  	theDefaultBrush = r;

  return TRUE;
}



BOOL Context::SelectFont(LOGFONT &plf,
		int iRotation, BOOL Scale )
{
  int Selected = -1;
  LOGFONT lf = plf;

  int iWidth = lf.lfWidth;
  int iHeight = lf.lfHeight;
  int Sign = iHeight>0 ? 1 : -1;

  if (Scale)
  {
	iHeight = theTransform.Scale(iHeight*Sign) * Sign;
	if (iWidth != 0)
	{
		iWidth = theTransform.Scale(iWidth);
	}
  }
  iRotation = theTransform.RotateDir(iRotation);

  if (theTransform.GetIsYUp())
  	iRotation = (iRotation>=2) ? 0 : -900;
  else
  	iRotation = (iRotation>=2) ? 0 : 900;

  	lf.lfHeight = iHeight;
	lf.lfEscapement = iRotation;
	lf.lfOutPrecision = OUT_TT_PRECIS;
	lf.lfQuality = PROOF_QUALITY;
	lf.lfWidth = iWidth;

  // Does this font already exist?
  font_map::iterator it = m_fonts.find( lf );

	CFont *r;

  // If it doesn't then create it
  if (it == m_fonts.end()) 
  {
	CFont *f = new CFont;
	f->CreateFontIndirect(&lf);
	m_fonts[lf] = f;

	r = dc->SelectObject(f);
  }
  else
  {
	  r = dc->SelectObject(it->second);
  }

  // Now select the font
  if (theDefaultFont == NULL)
  	theDefaultFont = r;

  return TRUE;
}




////// The drawing primitives //////

void Context::PolyBezier(POINT *pts,int Size)
{
  // Make a new array to hold the points
  POINT *Np = new POINT[Size];

  // Convert each of the points before calling PolyBezier
  for (int lp=0;lp!=Size;lp++)
	Np[lp] = theTransform.Scale(pts[lp]);

  // The PolyBezier function must be emulated in Win32s or for Metafiles
  if (IsMetaFile()) {
	if (Size >= 4) {
  		float x, y;
  		float t, t2, t3, a, b, c;
  		dc->MoveTo((int)Np[0].x,(int)Np[0].y);
  
  		for (t=0.0f; t<=1.0f; t += 0.02f) {
   			t2 = t * t;
   			t3 = t2 * t;
   			a = 1 - 3*t + 3*t2 - t3;
   			b = 3*(t - 2*t2 + t3);
   			c = 3*(t2 - t3);
   			x = a * Np[0].x
       			+ b * Np[1].x
       			+ c * Np[2].x
       			+ t3 * Np[3].x;
   			y = a * Np[0].y
       			+ b * Np[1].y
       			+ c * Np[2].y
       			+ t3 * Np[3].y;
   			dc->LineTo((int)x, (int)y);
  		}
 	} else {
  		dc->Polyline(Np, Size);
 	}
  } else {
#ifndef _cWIN16_
  ::PolyBezier(dc->m_hDC,Np,Size);
#endif
  }

  delete Np;
}




void Context::Arc( LPCRECT lpRect, POINT ptStart, POINT ptEnd )
{
 	CPoint Recta = theTransform.Scale(CPoint(lpRect->left ,lpRect->top));
 	CPoint Rectb = theTransform.Scale(CPoint(lpRect->right,lpRect->bottom));
 	CPoint Start = theTransform.Scale(ptStart);
 	CPoint End   = theTransform.Scale(ptEnd);

	dc->Arc(CRect(Recta.x,Recta.y,Rectb.x,Rectb.y), Start, End);
}



void Context::QuaterArc(CPoint a, CPoint b)
{
	int Width  = (a.x - b.x);
	int Height = (a.y - b.y);

	CRect theRect = CRect(b.x - Width, a.y - Height, b.x + Width, a.y + Height);
	theRect = theTransform.Scale(theRect);

	b = theTransform.Scale(b);
	a = theTransform.Scale(a);

	CPoint Mid = CPoint(theRect.left + (theRect.right  - theRect.left)/2,
					    theRect.top +  (theRect.bottom - theRect.top )/2 );

	double PI = 3.14159;

	double ra = atan2((double)(a.x - Mid.x), (double)(a.y- Mid.y));
	double rb = atan2((double)(b.x - Mid.x), (double)(b.y- Mid.y));
	double diff = ra - rb;

	if (diff < (-PI) || (diff > 0.0 && diff < PI )) {
		CPoint hold = a;
		a = b;
		b = hold;
	}

	if (theTransform.GetIsYUp()) {
		CPoint hold = a;
		a = b;
		b = hold;
	}

	dc->Arc(theRect, a, b);
}


void Context::Polyline(pointCollection &points, CPoint offset, FillStyle *pStyle)
{ 
	CPoint *lpPoints = new CPoint[ points.size() ];

	pointCollection::iterator it = points.begin();
	int n = 0;
	while (it != points.end())
	{
		lpPoints[n] = theTransform.Scale( *it + offset );
		++ it;
		++ n;
	}

	if (!pStyle || n == 2)
	{
		dc->Polyline( lpPoints, n );
	}
	else
	{
		SelectBrush( pStyle );
		dc->SetPolyFillMode( WINDING );
		dc->Polygon( lpPoints, n );
	}

	delete[] lpPoints;
}


////// The text drawing functions //////
//
// On entry the dc should contain the text you wish to display
// Over-bars will be added automatically
// This function will either display the text, or a greyed box

// Calculate the extent of the text on the DC
CSize Context::GetTextExtentI(CString text)
{
  int len=text.GetLength(),lp,lpi=0;
  char TextString[STRLEN];

  dc->SetTextAlign(TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);

  // seperate the strings into the overbars and text
  for (lp=0;lp!=len;lp++)
	if (text[lp]!='`')
		TextString[lpi++]=text[lp];

  TextString[lpi]=0;

  return dc->GetTextExtent(TextString,lpi);
}


// Display the text on the screen!
void Context::TextOut(CString text,CPoint a,paint_options options,int dir)
{
  int len=text.GetLength(),lp,lpi=0;
  char BarString[STRLEN],TextString[STRLEN];

  int oldalign = dc->SetTextAlign(TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);

  a = theTransform.Scale(a);
  dir = theTransform.RotateDir(dir);

  // seperate the strings into the overbars and text
  BarString[0]=' ';
  for (lp=0;lp!=len;lp++) {
	BarString[lp+1]=' ';
	if (text[lp]=='`')
		BarString[lpi]='_';
	else {
		TextString[lpi]=text[lp];
		lpi++;
	}
  }
  BarString[lpi]=0;
  TextString[lpi]=0;

  CSize size=dc->GetTextExtent(TextString,lpi);

  // move the positioning to follow the rotations
  if (dir==1)
	a=CPoint(a.x,a.y+size.cx);
  if (dir==2)
	a=CPoint(a.x-size.cx,a.y);
  


  	dc->SetROP2(R2_COPYPEN);
	dc->SetBkMode(TRANSPARENT);

	dc->TextOut(a.x,a.y,TextString);

	// Now add bars to the text
	CSize sz;
	char Last=' ';
	CPen *oldPen = NULL;
	
	if (theDefaultPen)
	{
		oldPen = dc->SelectObject( theDefaultPen );
	}

	SelectPen(PS_SOLID,1,cSELECT);
	if (dir>=2)
		dc->MoveTo(a.x,a.y-size.cy);
	else
		dc->MoveTo(a.x-size.cy,a.y);

	int last=0,thisx=a.x;
	for (lp=0;lp!=lpi;lp++) {
		if (Last!=BarString[lp]) {
			sz=dc->GetTextExtent(TextString,lp);
			if (BarString[lp]=='_') {
				if (dir>=2)
					dc->MoveTo(a.x+sz.cx,a.y-size.cy);
				else
					dc->MoveTo(a.x-size.cy,a.y-sz.cx);
			} else {
				if (dir>=2)
					dc->LineTo(a.x+sz.cx,a.y-size.cy);
				else
					dc->LineTo(a.x-size.cy,a.y-sz.cx);
			}
			Last=BarString[lp];
		}
	}
	if (Last!=' ') {
		if (dir>=2)
			dc->LineTo(a.x+size.cx,a.y-size.cy);
		else
			dc->LineTo(a.x-size.cy,a.y-size.cx);
	}

	if (oldPen)
	{
		dc->SelectObject( oldPen );
	}

  dc->SetTextAlign(oldalign);
}


// Text out with fixed width.
void Context::TextOut(int x,int y, int width, const char *t)
{
	CPoint r = theTransform.Scale(CPoint(x,y)); 
	int s_width = theTransform.Scale( width );
	int s_height = dc->GetTextExtent( t ).cy;

	CRect rect(r.x ,r.y - s_height,r.x+s_width,r.y );

	int old = dc->SetTextAlign( DT_TOP | DT_LEFT );
	dc->DrawText(t,rect,DT_SINGLELINE | DT_RIGHT | DT_TOP  );
	dc->SetTextAlign( old );
}



CRect Context::GetUpdateRegion()
{
  TransformSnap snap;

  snap.SetGridSnap(FALSE);

  CRect rect;
  rect = (((CPaintDC *)dc)->m_ps).rcPaint;

  CPoint Start = theTransform.DeScale(snap,CPoint(rect.left,rect.top));
  CPoint End = theTransform.DeScale(snap,CPoint(rect.right,rect.bottom));

  return CRect(Start.x,Start.y,End.x,End.y);
}
 
