/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of composite objects (symbols)

#include "stdafx.h"
#include "TinyCad.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include <math.h>
#include "JunctionUtils.h"

// The Method Class //

// NB a = offset of the bottom left hand corner (used to define position)
//    b = offset of the top right hand corner (used to define size)

// The constructor
CDrawMethod::CDrawMethod(CTinyCadDoc *pDesign,hSYMBOL symbol)
: CDrawingObject( pDesign )
{
  rotate=0;

  m_Symbol = symbol;

  CSymbolRecord *newSymbol = GetSymbolData();
  SymName=newSymbol->name;
  
  a = CPoint(0,0);
  b = GetTr();

  m_fields.resize( 2 );

  m_fields[Name].m_value=newSymbol->name;
  m_fields[Name].m_description="Name";
  m_fields[Name].m_show= newSymbol->name_type == default_show;
  m_fields[Name].m_position = CPoint(0,0);
  m_fields[Name].m_type = newSymbol->name_type;

  m_fields[Ref].m_value=newSymbol->reference;
  m_fields[Ref].m_description="Ref";
  m_fields[Ref].m_show=newSymbol->ref_type == default_show;
  m_fields[Ref].m_position = CPoint(0,0);
  m_fields[Ref].m_type = newSymbol->ref_type;

  // Now load in the other fields
  for (int i = 0; i < newSymbol->fields.size(); i++)
  {
	  CField f;
	  f.m_description = newSymbol->fields[i].field_name;
	  f.m_value = newSymbol->fields[i].field_default;
	  f.m_type = newSymbol->fields[i].field_type;
	  f.m_position = CPoint(0,0);
	  f.m_show = f.m_type == default_show;

	  m_fields.push_back( f );
  }

  rotate=0;
  part=0;

  segment=1;

  Undo_Level=0;
  Delete_Level=-1; 

  scaling_x = 1.0;
  scaling_y = 1.0;
  can_scale = FALSE;


  NewRotation();

}


// Tag the resources being used by this symbol
void CDrawMethod::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(fPIN);
	m_pDesign->GetOptions()->TagSymbol( m_Symbol );

	CPoint tr;
	drawingCollection method;
	ExtractSymbol( tr, method );

	drawingIterator it = method.begin();
	while (it != method.end()) 
	{
		(*it)->TagResources();
		++ it;
	}

}



// Look for a seach string in the object
CString CDrawMethod::Find(const char *theSearchString)
{
  CString HoldString;

  for (int lp =0;lp<m_fields.size();lp++) 
  {
  	HoldString = m_fields[lp].m_value;
  	HoldString.MakeLower();

  	// Now look for the search string in this string
	if (HoldString.Find(theSearchString) != -1)
  		return m_fields[lp].m_value;
  }

  return "";
}



// Rotate this object about a point
void CDrawMethod::Rotate(CPoint p,int ndir)
{
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		rotate ^= (rotate&4);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  // Set the a and b co-ords to their correct arrangements
  CPoint la = a;

  a = CPoint(max(la.x,b.x),max(la.y,b.y));
  b = CPoint(min(la.x,b.x),min(la.y,b.y));


  rotate = DoRotate(rotate&3,ndir) | (rotate&4);

  NewRotation();
}


int CDrawMethod::DoRotate(int olddir,int newdir)
{
  //     New rotation=> 2  3  4  
  //				    Current dir ..\/..
  int table[] = {	2, 3, 0, 		// 0 (Up)
			3, 2, 1,		// 1 (Down)
			1, 0, 3,		// 2 (Left)
			0, 1, 2,		// 3 (Right)
		};

  return table[(newdir-2) + olddir*3];
}


// Another Constructor
CDrawMethod::CDrawMethod(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
	m_Symbol = 0;
  rotate=0;

  SymName="";
  
  a=b=CPoint(0,0);

  m_fields.resize( 3 );

  m_fields[Name].m_value="";
  m_fields[Name].m_description="Name";
  m_fields[Name].m_show=TRUE;
  m_fields[Name].m_position = CPoint(0,0);
  m_fields[Name].m_type = default_show;

  m_fields[Other].m_value="";
  m_fields[Other].m_description="Other";
  m_fields[Other].m_show=FALSE;
  m_fields[Other].m_position = CPoint(0,0);
  m_fields[Other].m_type = default_hidden;

  m_fields[Ref].m_value="";
  m_fields[Ref].m_description="Ref";
  m_fields[Ref].m_show=TRUE;
  m_fields[Ref].m_position = CPoint(0,0);
  m_fields[Ref].m_type = default_show;

  rotate=3;
  part=0;

  segment=0;

  Undo_Level=0;
  Delete_Level=-1;

  scaling_x = 1.0;
  scaling_y = 1.0;
  can_scale = FALSE;
}


BOOL CDrawMethod::ExtractSymbol( CPoint &tr, drawingCollection &method )
{
	if (!GetSymbolData()->GetMethod( tr, method ))
	{
		return FALSE;
	}

  tr = CPoint(tr.x * scaling_x, tr.y * scaling_y );

  // Calculate the bounding box
  if ((rotate&3)<2)
	b=CPoint(tr.x+a.x,tr.y+a.y);
  else
	b=CPoint(tr.y+a.x,tr.x+a.y);

  return TRUE;
}

BOOL CDrawMethod::IsNoSymbol()
{
	return m_Symbol == 0;
}


// Get the method to draw this symbol
CDesignFileSymbol *CDrawMethod::GetSymbolData()
{
	if (IsNoSymbol())
	{
		// Try looking up this symbol
		GetSymbolByName();
	}
	
	return m_pDesign->GetOptions()->GetSymbol( m_Symbol );
}


// Get the reference value
int CDrawMethod::GetRefVal()
{
  return atoi(GetRef().Mid(GetRef().FindOneOf("0123456789")));
}

// Set the reference value
void CDrawMethod::SetRefVal(int value)
{
  char Buffer[10];

  _itoa(value,Buffer,10);
  m_fields[Ref].m_value = GetSymbolData()->reference;
  m_fields[Ref].m_value 
	  = m_fields[Ref].m_value.Left(m_fields[Ref].m_value.ReverseFind('?')) + Buffer;
}



BOOL CDrawMethod::CanEdit()
{
  return TRUE;
}

CString CDrawMethod::GetName() const
{
  return m_fields[Name].m_value;
}


void CDrawMethod::BeginEdit(BOOL re_edit)
{
  g_EditToolBar.m_MethodEdit.Open(m_pDesign,this);
}

void CDrawMethod::EndEdit()
{
  RButtonDown(CPoint(0,0), CPoint(0,0));
  g_EditToolBar.m_MethodEdit.Close();
}


CString CDrawMethod::GetField(int which)
{
  CString r = m_fields[which].m_value;

  if (which == Ref && GetSymbolData()->ppp!=1) {
	char Apart[2];
	if (GetSymbolData()->ppp>1) {
		Apart[0]=part + 'A';
		Apart[1]=0;
	} else
		Apart[0]=0;
	r += Apart;
  }

  return r;
}

void CDrawMethod::GetSymbolByName()
{
  CLibraryFileSymbol *theSymbol = static_cast<CTinyCadApp*>(AfxGetApp())->Extract(SymName);
  if (theSymbol != NULL)
  {
	m_Symbol = m_pDesign->GetOptions()->AddSymbol( theSymbol->GetDesignSymbol(m_pDesign) );
  }
  else
  {
	  // Use the no symbol...
	  m_Symbol = 0;
  }
}


void CDrawMethod::OldLoad(CArchive &archive)
{
  BYTE Shows;

  m_fields.resize(3);

  archive >> SymName 
	  >> m_fields[Name].m_value 
	  >> m_fields[Other].m_value 
	  >> m_fields[Ref].m_value;
  m_fields[Name].m_position  = ReadPoint(archive);
  m_fields[Other].m_position = ReadPoint(archive);
  m_fields[Ref].m_position   = ReadPoint(archive);

  m_fields[Other].m_description="Other";
  m_fields[Other].m_type = default_hidden;

  archive >> Shows >> rotate;
  a = ReadPoint(archive);
  archive >> part;

  m_fields[Name].m_show  = (Shows & 1)!=0;
  m_fields[Other].m_show = (Shows & 2)!=0;
  m_fields[Ref].m_show   = (Shows & 4)!=0;

  GetSymbolByName();  

  segment = 0;

  // Now get the calculation of the bounding box
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );
}


void CDrawMethod::OldLoad2(CArchive &archive)
{
  BYTE Shows;
  CString attrib;

  m_fields.resize( 3 );

  archive >> SymName 
	  >> m_fields[Name].m_value 
	  >> m_fields[Other].m_value 
	  >> m_fields[Ref].m_value
	  >> attrib;
  m_fields[Name].m_position  = ReadPoint(archive);
  m_fields[Other].m_position = ReadPoint(archive);
  m_fields[Ref].m_position   = ReadPoint(archive);

  m_fields[Other].m_description="Other";
  m_fields[Other].m_type = default_hidden;

  archive >> Shows >> rotate;
  a = ReadPoint(archive);
  archive >> part;

  m_fields[Name].m_show  = (Shows & 1)!=0;
  m_fields[Other].m_show = (Shows & 2)!=0;
  m_fields[Ref].m_show   = (Shows & 4)!=0;

  GetSymbolByName();

  segment = 0;

  // Now get the calculation of the bounding box
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );
}

void CDrawMethod::OldLoad3(CArchive &archive)
{
  BYTE Shows;

  CString attrib;

  m_fields.resize(3);

  archive >> m_Symbol;
  archive >> m_fields[Name].m_value 
	  >> m_fields[Other].m_value 
	  >> m_fields[Ref].m_value
	  >> attrib;
  m_fields[Name].m_position  = ReadPoint(archive);
  m_fields[Other].m_position = ReadPoint(archive);
  m_fields[Ref].m_position   = ReadPoint(archive);

  m_fields[Other].m_description="Other";
  m_fields[Other].m_type = default_hidden;

  archive >> Shows >> rotate;
  a = ReadPoint(archive);
  archive >> part;

  // Read in the special other text
  std::vector<CSymbolField> fields;
  WORD special_text_count;
  archive >> special_text_count;
  for (int i = 0; i < special_text_count; i++)
  {
	  CSymbolField sf;
	  sf.Load( archive );
	  fields.push_back( sf );
  }

  for (i = 0; i < special_text_count; i++)
  {
	  if (fields[i].field_name == "SCALING_X")
	  {
		  scaling_x = atof(fields[i].field_value);
	  }
	  else if (fields[i].field_name == "SCALING_Y")
	  {
		  scaling_y = atof(fields[i].field_value);
	  }
  }


  m_fields[Name].m_show  = (Shows & 1)!=0;
  m_fields[Other].m_show = (Shows & 2)!=0;
  m_fields[Ref].m_show   = (Shows & 4)!=0;

  m_Symbol = m_pDesign->GetOptions()->GetNewSymbolNumber( m_Symbol );

  segment = 0;


  // Now get the calculation of the bounding box
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );
}

void CDrawMethod::Save(CArchive &archive)
{
  int load_version = 1;

  // Save the method revision number
  // (This supercedes the need for MethodEx, Ex2, Ex3 etc..)
  archive << load_version;

  // Now write out our symbol number
  archive << m_Symbol;

  // Our location
  WritePoint(archive,a);

  // Part in package...
  archive << part;

  // Our rotation
  archive << rotate;

  // Scaling
  archive << can_scale << scaling_x << scaling_y;

  // Now read in the fields
  int field_size = m_fields.size();;
  archive << field_size;
  for (int i = 0; i < m_fields.size(); i++)
  {
	  CField &f = m_fields[i];

	  int t = static_cast<int>(f.m_type);
	  archive << t;

	  archive 
		<< f.m_description
		<< f.m_value
		<< f.m_show
		<< f.m_position;
  }

}


void CDrawMethod::Load(CArchive &archive)
{
  int load_version;

  // Pull in the method revision number
  // (This supercedes the need for MethodEx, Ex2, Ex3 etc..)
  archive >> load_version;

  // Now write out our symbol number
  archive >> m_Symbol;

  // Our location
  a = ReadPoint(archive);

  // Part in package...
  archive >> part;

  // Our rotation
  archive >> rotate;

  // Scaling
  archive >> can_scale >> scaling_x >> scaling_y;

  // Now read in the fields
  int field_size;
  archive >> field_size;
  m_fields.resize( field_size );
  for (int i = 0; i < m_fields.size(); i++)
  {
	  CField &f = m_fields[i];

	  int t;
	  archive >> t;
	  f.m_type = static_cast<SymbolFieldType>(t);

	  archive 
		>> f.m_description
		>> f.m_value
		>> f.m_show
		>> f.m_position;
  }


  // Use this data to locate our symbol
  m_Symbol = m_pDesign->GetOptions()->GetNewSymbolNumber( m_Symbol );

  segment = 0;

  // Now get the calculation of the bounding box
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );
}



inline void swap(int &a,int &b) { int sp; sp=a; a=b; b=sp; }

double CDrawMethod::DistanceFromPoint( CPoint p )
{
	// If definately inside this item then return 0
	if (IsInside( p.x,p.x, p.y,p.y ))
	{
		return 0.0;
	}

	// Ok, try a looser fit (and return 20 if we do fit...)
   CPoint tr;
   drawingCollection method;
   ExtractSymbol( tr, method );

  // Use fast cut-off to see if the bounding box is inside
  // the intersection box
  if ( !((a.x<p.x && b.x<=p.x) || (a.x>p.x && b.x>p.x)
    || (a.y<p.y && b.y<=p.y) || (a.y>p.y && b.y>p.y))) 
  {
	return 15.0;
  }


  return 100.0;
}



BOOL CDrawMethod::IsInside(int left,int right,int top,int bottom)
{
   CPoint tr;
   drawingCollection method;
   ExtractSymbol( tr, method );


  // Use fast cut-off to see if the bounding box is inside
  // the intersection box
  if ( !((a.x<left && b.x<=left) || (a.x>right && b.x>right)
    || (a.y<top && b.y<=top) || (a.y>bottom && b.y>bottom))) {
	int nl,nr,nt,nb;

	// Translate the intersection box
	double sx = scaling_x;
	double sy = scaling_y;
	if ((rotate&3) >= 2)
	{
		sy = scaling_x;
		sx = scaling_y;
	}

	nl = (left-a.x) / sx;
	nr = (right-a.x) / sx;

	nt = (top-a.y) / sy;
	nb = (bottom-a.y) / sy;

	// Rotate the intersection box
	switch (rotate&3) {
		case 1:		// Down
			swap(nt,nb);
			nt=-(nt-tr.y/sy);
			nb=-(nb-tr.y/sy);
			break;
		case 2:		// Left
			swap(nt,nl);
			swap(nb,nr);
			break;
		case 3:		// Right
			swap(nt,nl);
			swap(nb,nr);
			swap(nt,nb);
			nt=-(nt-tr.y/sx);
			nb=-(nb-tr.y/sx);
			break;
	}

	// Mirror the intersection box
	if ((rotate&4)!=0) {
		if ((rotate&3) >= 2)
		{
			nl=-nl+tr.x/sy;
			nr=-nr+tr.x/sy;
			swap(nl,nr);
		}
		else
		{
			nt=-nt+tr.y/sy;
			nb=-nb+tr.y/sy;
			swap(nb,nt);
		}
	}

	// Select the correct part for this device
	int hold = m_pDesign->GetPart();
	m_pDesign->SetPart( part );

	// Search each element until one is inside
	drawingIterator it = method.begin();
	while (it != method.end())
	{
		if ((*it)->IsInside(nl,nr,nt,nb))
		{
			break;
		}

		++ it;
	}


	m_pDesign->SetPart( hold );

	if (it != method.end())
		return TRUE;
  }

  int lp;
  for (lp=0;lp < m_fields.size();lp++) {
	if (!m_fields[lp].m_show)
	{
		continue;
	}

	CPoint p1,p2;

	CSize size=m_pDesign->GetTextExtent(GetField(lp),fPIN);

	p1 = CPoint(m_fields[lp].m_position.x+a.x,m_fields[lp].m_position.y+a.y);
	p2 = CPoint(m_fields[lp].m_position.x+a.x+size.cx,m_fields[lp].m_position.y+a.y-size.cy);
	if ( !( (p1.x<left && p2.x<=left) || (p1.x>right && p2.x>right)
	     || (p1.y<top && p2.y<=top) || (p1.y>bottom && p2.y>bottom)) )
		return TRUE;

  }

  return FALSE;
}


ObjType CDrawMethod::GetType()
{
  return xMethodEx3;
}


void CDrawMethod::Move(CPoint p)
{
  // r is the relative displacement to move
  CPoint r;

  r.x=p.x-a.x;
  r.y=p.y-a.y;

  if (r.x!=0 || r.y!=0) 
  {
	Display();

	a=p;
	b=CPoint(b.x+r.x,b.y+r.y);

	segment=0;
	Display();
  }
}


void CDrawMethod::LButtonDown(CPoint p, CPoint)
{
  Display();
  Move(p);
  Store();

  CJunctionUtils j(m_pDesign);
  j.AddObjectToTodo( this );
  j.CheckTodoList();
 
  Display();	// Write to screen
}



void CDrawMethod::MoveField(int w, CPoint r)
{
  Display();

  if (w >= 1000)
  {
	  m_fields[w - 1000].m_position.x += r.x;
	  m_fields[w - 1000].m_position.y += r.y;
	  Display();
  }
  else
  {
	CRect rect( a.x, a.y, b.x, b.y );

	switch (w)
	{
	case CRectTracker::hitTopLeft:
		rect.left += r.x;
		rect.top += r.y;
		break;
	case CRectTracker::hitTopRight:
		rect.right += r.x;
		rect.top += r.y;
		break;
	case CRectTracker::hitBottomRight:
		rect.right += r.x;
		rect.bottom += r.y;
		break;
	case CRectTracker::hitBottomLeft:
		rect.left += r.x;
		rect.bottom += r.y;
		break;	
	case CRectTracker::hitTop:
		rect.top += r.y;
		break;		
	case CRectTracker::hitRight:
		rect.right += r.x;
		break;		
	case CRectTracker::hitBottom:
		rect.bottom += r.y;
		break;		
	case CRectTracker::hitLeft:
		rect.left += r.x;
		break;		
	}

	// Determine scaling in the x and y
	CPoint tr;
	scaling_x = 1.0;
	scaling_y = 1.0;
	drawingCollection method;
	ExtractSymbol( tr, method );
	scaling_x = fabs(rect.Width()) / fabs( a.x - b.x );
	scaling_y = fabs(rect.Height()) / fabs( a.y - b.y );

	if ((rotate&3) >= 2)
	{
		double q = scaling_x;
		scaling_x = scaling_y;
		scaling_y = q;
	}

	a.x = rect.left;
	a.y = rect.top;
	b.x = rect.right;
	b.y = rect.bottom;


	Display();
  }
}


int CDrawMethod::IsInsideField(CPoint p)
{
  int lp;
  CSize size;

	if (can_scale)
	{
		CRectTracker tracker( CRect(a.x,a.y,b.x,b.y), CRectTracker::dottedLine | CRectTracker::resizeOutside  );
		int r = tracker.HitTest( p );

		if (r != CRectTracker::hitNothing && r != CRectTracker::hitMiddle)
		{
			return r;
		}
	}

  // Convert co-ords to symbol space
  p.x = p.x -a.x;
  p.y = p.y -a.y;

  for (lp=0;lp < m_fields.size();lp++) 
  {
	if (!m_fields[lp].m_show)
	{
		continue;
	}

	size=m_pDesign->GetTextExtent(GetField(lp),fPIN);

	if (p.x>=m_fields[lp].m_position.x && p.x<=m_fields[lp].m_position.x+size.cx
	 && p.y<=m_fields[lp].m_position.y && p.y>=m_fields[lp].m_position.y-size.cy)
		break;
  }

  if (lp<m_fields.size())
	return lp + 1000;
  else
	return -1;
}


// The space between the method and text
#define SPACING 5


void CDrawMethod::NewRotation()
{
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );

  // Calculate the bounding box
  if ((rotate&3)<2)
	b=CPoint(tr.x+a.x,tr.y+a.y);
  else
	b=CPoint(tr.y+a.x,tr.x+a.y);


  // Calculate the text positions
  int TextHeight=sPIN;
  if (TextHeight <0)
  	TextHeight = -TextHeight;
  int NextY=TextHeight;

  if (m_fields[Name].m_show) {
	if ((rotate&3)<2) {
		// Display to right hand side if display up or down
		m_fields[Name].m_position=CPoint(SPACING,b.y+NextY-a.y);
		NextY+=TextHeight+SPACING/2;
	} else
		// Display above if left or right
		m_fields[Name].m_position=CPoint(b.x-a.x,b.y-a.y);
  }

  CPoint p;

	if ((rotate&3)<2)
	{
		// Display to right hand side if display up or down
		p=CPoint(SPACING,b.y+NextY-a.y);
	}
	else
	{
		// Display below if left or right
		p=CPoint(b.x-a.x,NextY);
	}

	// First place the shown fields
	for (int i = 0; i < m_fields.size(); i++)
	{
		// Don't do the name
		if (i==Name)
		{
			continue;
		}

		// Place the other fields
		if (m_fields[i].m_show) 
		{
			m_fields[i].m_position = p;
			p.y += TextHeight+SPACING/2;
		}
	}

	// Now place the hidden fields...
	for (i = 0; i < m_fields.size(); i++)
	{
		// Don't do the name
		if (i==Name)
		{
			continue;
		}

		// Place the other fields
		if (!m_fields[i].m_show && m_fields[i].m_type != always_hidden) 
		{
			m_fields[i].m_position = p;
			p.y += TextHeight+SPACING/2;
		}
	}
}

CPoint CDrawMethod::GetTr()
{ 
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );
	return tr; 
}

void CDrawMethod::Display( BOOL erase )
{
	// Invalidate the symbol
	CRect r( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r, erase, 10 );

  // Now invalidate our text
  for (int lp=0;lp < m_fields.size(); lp++)
  {
	if (m_fields[lp].m_show)
	{
		CSize sz = m_pDesign->GetTextExtent( GetField(lp), fPIN );
		r.left = m_fields[lp].m_position.x+a.x;
		r.top = m_fields[lp].m_position.y+a.y;
		r.right = r.left + sz.cx;
		r.bottom = r.top - sz.cy;
		m_pDesign->InvalidateRect( r, erase, 5 );
	}
  }
}

void CDrawMethod::Paint(Context &dc,paint_options options)
{
  CPoint tr;
  drawingCollection method;
  ExtractSymbol( tr, method );

  // Now rescale this device...
  double old_scaling_x;
  double old_scaling_y;
  (dc.theTransform).GetScaling( rotate, old_scaling_x, old_scaling_y );
  (dc.theTransform).SetScaling( rotate, scaling_x, scaling_y );

  // Move the object about by using the grid offset!
  // (that'll have to do until I think of a better method)
  CPoint oldpos= (dc.theTransform).SetTRM(a,tr,rotate);


  // Select the correct part for this device
  int OldPart = m_pDesign->GetPart();
  m_pDesign->SetPart(part);


  drawingIterator it = method.begin();
  while (it != method.end()) 
  {
	(*it)->Paint(dc,options);
	++ it;
  }

  (dc.theTransform).SetScaling( rotate, old_scaling_x, old_scaling_y );
  (dc.theTransform).EndTRM(oldpos);
  m_pDesign->SetPart(OldPart);



  // Now display the text (if necessary)
  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(fPIN),2);
  switch (options)
  {
  case draw_selected:
	  dc.SetTextColor(cPIN);
	  break;
  case draw_selectable:
  	  dc.SetTextColor(cPIN_CLK);
	  break;
  default:
	  dc.SetTextColor(cPIN);
  }

  for (int lp=0;lp < m_fields.size();lp++)
  {
	if (m_fields[lp].m_show)
		dc.TextOut(GetField(lp),
			CPoint(m_fields[lp].m_position.x+a.x,
				m_fields[lp].m_position.y+a.y),options);
  }

}




// Store this method in the drawing
CDrawingObject* CDrawMethod::Store()
{
  CDrawMethod *NewObject;

  NewObject = new CDrawMethod(m_pDesign);

  *NewObject=*this;
  
  m_pDesign->Add(NewObject);

  return NewObject;
}



void CDrawMethod::PaintHandles( Context &dc )
{
	if (can_scale)
	{
		// Put some handles around this object
		CRect r(a.x,a.y,b.x,b.y);

		CRectTracker tracker( dc.theTransform.Scale(r), CRectTracker::dottedLine | CRectTracker::resizeOutside  );
		tracker.Draw( dc.dc );
	}
}

int CDrawMethod::SetCursorEdit( CPoint p )
{
	CRect r(a.x,a.y,b.x,b.y);
	r.NormalizeRect();
	CRectTracker tracker( r, CRectTracker::dottedLine | CRectTracker::resizeOutside  );
	int q = tracker.HitTest( p );

	if (q == 8 || !can_scale)
	{
		q = DistanceFromPoint( p ) <= 20;
		return q ? 11 : -1;
	}

	return q;
}


// Add/remove the next references
void CDrawMethod::AddReference( int min_ref )
{
	// Search the document for a gap in our references
	// 
	CString our_reference = GetSymbolData()->reference;
	int last_ref_number = min_ref - 1;
	int last_ppp_number = 0;

	drawingIterator it = m_pDesign->GetDrawingBegin();
	while (it != m_pDesign->GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;

		if (pointer->GetType() == xMethodEx3) 
		{
			CDrawMethod *pMethod = static_cast<CDrawMethod*>( pointer );

			if (pMethod->GetSymbolData()->reference == our_reference)
			{
				// This is the same reference, so we keep track
				// of the maximum value
				int ref_number = pMethod->GetRefVal();
				int ppp_number = pMethod->GetSubPart();
				if (    last_ref_number < ref_number
					|| (last_ref_number == ref_number && last_ppp_number < ppp_number) )
				{
					last_ref_number = ref_number;
					last_ppp_number = ppp_number;
				}
			}
		}

		++ it;
	}

	// Now we know what the last value was, so set
	// us to the next value
	last_ppp_number ++;
	if (last_ppp_number >= GetSymbolData()->ppp)
	{
		last_ppp_number = 0;
		last_ref_number ++;
	}

	SetRefVal(last_ref_number);
	SetPart(last_ppp_number);
}

void CDrawMethod::RemoveReference()
{
	SetRef(GetSymbolData()->reference);
	SetPart(0);
}

