/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "layer.h"
#include "option.h"
#include "special.h"
#include "revision.h"
#include "TinyCad.h"

extern Option theOptions;


void CTinyCadDoc::Save(CArchive&theArchive, BOOL Details , BOOL SaveSelect)
{
	Save( theArchive, m_drawing, Details, SaveSelect );
}

////// Save a design to current position in a file //////

void CTinyCadDoc::Save(CArchive &theArchive, drawingCollection &drawing, BOOL Details, BOOL SaveSelect, BOOL SaveResources )
{
  // Write the objects to the file
  try
  {
	// Write the header to identify this file
	HeaderStamp(theArchive);

	// If necessary write the header stamp
	if (Details) 
	{
		LONG MaxXLong = MaxXSize, MaxYLong = MaxYSize;
		BYTE Shows = 0;
		if (ShowDetails)
		{
			Shows |= 1;
		}
		if (ShowRulers)
		{
			Shows |= 2;
		}

		theArchive << (BYTE)xDesignInformation2;
		theArchive << (LONG)MaxXLong << MaxYLong
			<< Title << Author << Revision << DocNumber << Organisation << Sheets << Shows << DateStr;
		m_snap.Save( theArchive );


	}


	if (SaveResources)
	{
		// Find out which resources are in use
		drawingIterator it = drawing.begin();
		while (it != drawing.end()) 
		{
			(*it)->TagResources();
			++ it;
		}

		// Save the resource details details
		theArchive << (BYTE)xFont;
		theOptions.SaveFonts(theArchive);
		theArchive << (BYTE)xLineStyle;
		theOptions.SaveStyles(theArchive);
		theArchive << (BYTE)xFillStyle;
		theOptions.SaveFillStyles(theArchive);
		theArchive << (BYTE)xMetaFiles;
		theOptions.SaveMetaFiles(theArchive);
	}

	// Only save the symbols if we are not saving
	// to a library or the header...
	if (Details)
	{
		theArchive << (BYTE)xSymbols;
		theOptions.SaveSymbols(theArchive);
	}

	if (Details)
	{
		theOptions.SaveOptions( theArchive );
	}

	drawingIterator it = drawing.begin();
	while (it != drawing.end()) 
	{
		CDrawingObject *obj = *it;

		if (obj->GetType() != xError
			&& 	(Details || !obj->IsConstruction()) 
			&&  (!SaveSelect || IsSelected( obj )) )
		{

			// Write the object tag
			theArchive << ((BYTE)(obj->GetType()));

			// Now save the actual object data
			obj->Save(theArchive);
		}

		++ it;
	}

  theArchive << (BYTE)xEndFile;

  }

  catch ( CFileException e) 
  {
	// Could not save the file properly
	Message(IDS_ABORTSAVE,MB_ICONEXCLAMATION);
  	return;
  }
}



// Write the header to a file
void HeaderStamp(CArchive &theArchive)
{
	CString name = NAME;
  theArchive << name  << (BYTE)FILE_REVISION << (WORD)FILE_MAGIC;
}


// Read back the header stamp and check it is OK to read this file
BOOL CheckHeaderStamp(CArchive &theArchive)
{
  BYTE tp=-1,revision;
  WORD magic;
  CString name;

  theArchive >> name >> revision >> magic;
  if (name!=NAME || magic!=FILE_MAGIC) {
	// Not written by this program
	Message(IDS_ABORTNOTMINE,MB_ICONEXCLAMATION);
	return FALSE;
  }

  if (revision>FILE_REVISION) {
	// Written by a more advanced version of this program
	Message(IDS_ABORTVERSION,MB_ICONEXCLAMATION);
	return FALSE;
  }

  return TRUE;
}



////// Load a design from a file, loaded design will be selected //////
BOOL CTinyCadDoc::ReadFile(CArchive &theArchive, BOOL Details, drawingCollection &drawing)
{
  try
  {
	// Save the old layer setting
	CDrawingObject *obj = NULL;
	BYTE tp=-1;

  	if (!CheckHeaderStamp(theArchive))
		return FALSE;

	while (tp!=xEndFile) {
		theArchive >> tp;
		switch (tp) {
			case xDesignInformation:
				// Only read this information if necessary
				if (Details) {
					CTime Date;
					LONG MaxXLong, MaxYLong;
					BYTE dummyb;
				  	theArchive >> MaxXLong >> MaxYLong
						>> Title >> Author >> Revision >> DocNumber >> Organisation >> Sheets >> dummyb >> Date;
					MaxXSize = MaxXLong;
					MaxYSize = MaxYLong;
					ShowDetails = dummyb;
					DateStr = Date.Format("%d %B %Y");

				} else {
					CString dummy;
					LONG dummyi;
					BYTE dummyb;
					CTime dummyt;
					theArchive >> dummyi >> dummyi
						>> dummy >> dummy >> dummy >> dummy >> dummy >> dummy >> dummyb >> dummyt;
				}
				obj = NULL;
				break;
			case xDesignInformation2:
				if (Details) {
					LONG MaxXLong, MaxYLong;
					BYTE Shows;
				  	theArchive >> MaxXLong >> MaxYLong
						>> Title >> Author >> Revision >> DocNumber >> Organisation >> Sheets >> Shows >> DateStr;
					MaxXSize = MaxXLong;
					MaxYSize = MaxYLong;
					ShowRulers = Shows & 2;
					ShowDetails = Shows & 1;
				} else {
					CString dummy;
					LONG dummyi;
					BYTE dummyb;
					theArchive >> dummyi >> dummyi
						>> dummy >> dummy >> dummy     >> dummy     >> dummy        >> dummy  >> dummyb >> dummy;
				}
				m_snap.Load( theArchive, Details );
				break;
			case xOptions:
				theOptions.LoadOptions( theArchive, Details );
				obj = NULL;
				break;
			case xFont:
				theOptions.LoadFonts(theArchive);
				obj = NULL;
				break;
			case xLineStyle:
				theOptions.LoadStyles(theArchive);
				obj = NULL;
				break;
			case xFillStyle:
				theOptions.LoadFillStyles(theArchive);
				obj = NULL;
				break;
			case xMetaFiles:
				theOptions.LoadMetaFiles(theArchive);
				obj = NULL;
				break;
			case xSymbols:
				theOptions.LoadSymbols(theArchive);
				obj = NULL;
				break;
			case xLayerTable:
				{
					// Read in the number of Layers in this list
					CString theNewName;
					WORD NumberOfResources;
					theArchive >> NumberOfResources;

					hRESOURCE OldResourceNumber = 0;

					while (NumberOfResources > 0) {
						theArchive >> OldResourceNumber;
						theArchive >> theNewName;
						NumberOfResources--;
					}
					obj = NULL;
				}
				break;
			case xRuler:
				obj = new CDrawRuler(this,FALSE);
				break;
			case xBus:
				obj = new CDrawLine(this,xBus);
				break;
			case xWire:
				obj = new CDrawLine(this,xWire);
				break;
			case xBusName:
				obj = new CDrawText(this,xBusName);
				break;
			case xBusNameEx:
				obj = new CDrawText(this,xBusNameEx);
				break;
			case xLabel:  	// The Old Label type
				obj = new CDrawText(this,xLabel);
				break;
			case xLabelEx:
				obj = new CDrawText(this,xLabelEx);
				break;
			case xLabelEx2:
				obj = new CDrawText(this,xLabelEx2);
				break;
			case xJunction:
				obj = new CDrawJunction(this);
				break;
			case xNoConnect:
				obj = new CDrawNoConnect(this);
				break;
			case xBusSlash:
				obj = new CDrawBusSlash(this);
				break;
			case xPower:
				obj = new CDrawPower(this);
				break;
			case xPin:
			case xPinEx:
				obj = new CDrawPin(this);
				break;
			case xLine:
				obj = new CDrawPolygon(this,xLine);
				break;
			case xLineEx:
				obj = new CDrawPolygon(this,xLineEx);
				break;
			case xLineEx2:
				obj = new CDrawPolygon(this,xLineEx2);
				break;
			case xPolygon:
				obj = new CDrawPolygon(this,xPolygon);
				break;
			case xDash:
				obj = new CDrawPolygon(this,xDash);
				break;
			case xText:		// The old text type
				obj = new CDrawText(this,xText);
				break;
			case xTextEx:
				obj = new CDrawText(this,xTextEx);
				break;
			case xTextEx2:
				obj = new CDrawText(this,xTextEx2);
				break;
			case xCircle:
			case xCircleEx:
			case xCircleEx2:
			case xCircleEx3:
				obj = new CDrawSquare(this,xCircleEx3);
				break;
			case xSquare:
			case xSquareEx:
			case xSquareEx2:
			case xSquareEx3:
				obj = new CDrawSquare(this,xSquareEx3);
				break;
			case xArc:
			case xArcEx:
			case xArcEx2:
				obj = new CDrawPolygon(this,static_cast<ObjType>(tp));
				break;
			case xMethod:		// The old method type
			case xMethodEx:
			case xMethodEx2:
			case xMethodEx3:
				obj = new CDrawMethod(this);
				break;
			case xTag:
				obj = new CDrawTag(this);
				break;
			case xMetaFile:
				obj = new CDrawMetaFile(this);
				break;
			default:
				obj = NULL;
				break;
		}
		
		if (obj!=NULL) 
		{
			// Provide special handling for old objects
			switch (tp) {
				case xSquare:
				case xSquareEx:
				case xCircle:
				case xCircleEx:
					((CDrawSquare *)obj)->OldLoad(theArchive,tp);
					break;
				case xCircleEx2:
				case xSquareEx2:
					((CDrawSquare *)obj)->OldLoad2(theArchive);
					break;
				case xMethod:
					((CDrawMethod *)obj)->OldLoad(theArchive);
					break;
				case xMethodEx:
					((CDrawMethod *)obj)->OldLoad2(theArchive);
					break;
				case xMethodEx2:
					((CDrawMethod *)obj)->OldLoad3(theArchive);
					break;
				case xPin:
					((CDrawPin *)obj)->OldLoad(theArchive);
					break;
				default:
					obj->Load(theArchive);
					break;
			}

			// Now add object to linked list
			drawing.insert( drawing.end(), obj );
		}
  	}

  }

  catch( CFileException&) 
  {
	Message(IDS_ABORTLOAD,MB_ICONEXCLAMATION);
  }


  return drawing.size() > 0;
}






char *TheObjects[] = {
	"Bus","Wire","BusName","Label","Junction","No Connect","Bus Entry","Power",
	"Pin","Line","Text","Ellipse","Rectangle","Arc","Symbol","Tag", "*" };


char *TheOperators[] = {
	"OBJECT", "TITLE", "AUTHOR", "ORGANISATION", "FILENAME", "REVISION", "DOCUMENT",
	"DATE", "SHEET", "UNITS", "LAYER", "POSITION", "TO", "DIRECTION",
	"VALUE", "STYLE" , "TYPE", "PART", "REF", "NAME", "OTHER", "ATTRIBUTE",
	"NUMBER", "ELEC", "SHOW", "*" };


// Read an operator from a string
int GetOp(char * &theBuffer)
{
	int c = 0;

	#define Sep(a) (a<=32 || a==',' || a=='\'')

	// Find the start of the operator
	while (Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;
	char *Start = theBuffer+c;

	// Find the end of the operator
	while (!Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;

	// Add a terminating char on the end of the operator if necessary
	theBuffer = theBuffer+c;
	if (theBuffer[0] != 0) {
		theBuffer[0] = 0;
		theBuffer++;
	}

	// Now convert this string into a number
	int r = 0;
	while (strcmp(Start,TheOperators[r]) != 0 && TheOperators[r][0] != '*')
		r++;

	return r;
}


// Read a string from a string
char *GetString(char * &theBuffer)
{
	int c = 0;

	// Find the start of the string
	while (theBuffer[c] != '\'' && theBuffer[c] != 0)
		c++;
	if (theBuffer[c]!=0)
		c++;
	char *Start = theBuffer+c;

	// Find the end of the string
	while (theBuffer[c]!='\'' && theBuffer[c] != 0)
		c++;

	// Add a terminating char on the end of the operator if necessary
	theBuffer = theBuffer+c;
	if (theBuffer[0] != 0) {
		theBuffer[0] = 0;
		theBuffer++;
	}

	return Start;
}


// Read an interger from a string
int GetInt(char * &theBuffer)
{
	int c = 0;

	#define Sep(a) (a<=32 || a==',' || a=='\'')

	// Find the start of the integer
	while (Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;
	char *Start = theBuffer+c;

	// Find the end of the string
	while (!Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;

	// Add a terminating char on the end of the operator if necessary
	theBuffer = theBuffer+c;
	if (theBuffer[0] != 0) {
		theBuffer[0] = 0;
		theBuffer++;
	}

	return atoi(Start);
}


// Read a CPoint from a string
CPoint GetCPoint(char * &theBuffer)
{

	double fx,fy;
	int c = 0;
	char *Start;
	
	// Get the first part of the cpoint
	while (Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;
	Start = theBuffer+c;

	// Find the comma or spaces separating the cpoint
	while (theBuffer[c]!=',' && !Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;

	theBuffer[c] = 0;
	fx = atof(Start);
	c++;
	

	// Find the next part of the CPoint
	while ((Sep(theBuffer[c]) || theBuffer[c]==',') && theBuffer[c] != 0)
		c++;
	Start = theBuffer+c;

	// Find the end of the string
	while (theBuffer[c]!='m' && theBuffer[c]!='"' && !Sep(theBuffer[c]) && theBuffer[c] != 0)
		c++;

	// Get the units this is in
	char Units = theBuffer[c];

	// Add a terminating char on the end of the operator if necessary
	theBuffer = theBuffer+c;
	if (theBuffer[0] != 0) {
		theBuffer[0] = 0;
		theBuffer++;
	}

	// Convert the second part into a float
	fy = atof(Start);

	// Now convert the units to mm
	switch (Units) {
		case 'm':	// mm
			theBuffer++;
			break;
		case '"':	// inches
			fy *= 25.4;
			fx *= 25.4;
			break;
	}


	// convert from mm to pixels
	fy *= (float)PIXELSPERMM;
	fx *= (float)PIXELSPERMM;

	return CPoint((int)fx,(int)fy);
}



