/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "tinycad.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"
#include "LibraryView.h"
#include "LibraryDoc.h"
#include "revision.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CLibraryView

IMPLEMENT_DYNCREATE(CLibraryView, CScrollView)

CLibraryView::CLibraryView()
{
}

CLibraryView::~CLibraryView()
{
}


BEGIN_MESSAGE_MAP(CLibraryView, CScrollView)
	//{{AFX_MSG_MAP(CLibraryView)
	ON_COMMAND(ID_SYMBOL_DELETESYMBOL, OnSymbolDeletesymbol)
	ON_COMMAND(ID_SYMBOL_EDITSYMBOL, OnSymbolEditsymbol)
	ON_COMMAND(ID_SYMBOL_RENAMESYMBOL, OnSymbolRenamesymbol)
	ON_UPDATE_COMMAND_UI(ID_SYMBOL_DELETESYMBOL, OnUpdateSymbolDeletesymbol)
	ON_UPDATE_COMMAND_UI(ID_SYMBOL_EDITSYMBOL, OnUpdateSymbolEditsymbol)
	ON_UPDATE_COMMAND_UI(ID_SYMBOL_RENAMESYMBOL, OnUpdateSymbolRenamesymbol)
	ON_COMMAND(ID_SYMBOL_NEWSYMBOL, OnSymbolNewsymbol)
	ON_WM_LBUTTONDOWN()
	ON_WM_CREATE()
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP

	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)

END_MESSAGE_MAP()


// The OnCreate function, called when the window is created
int CLibraryView::OnCreate( LPCREATESTRUCT q )
{
	int r = CView::OnCreate( q );

	if (r == 0)
	{
		  ClipboardFormat = RegisterClipboardFormat(CLIPBOARD_SY_FORMAT);
	}

	return r;
}

BOOL CLibraryView::IsClipboardAvailable()
{
	return ::IsClipboardFormatAvailable(ClipboardFormat);
}


/////////////////////////////////////////////////////////////////////////////
// CLibraryView drawing

void CLibraryView::OnDraw(CDC* pDC)
{
	CLibraryDoc * pDoc = GetDocument();

	CLibraryFile::symbolCollection::iterator it;
	
	if (pDC->IsPrinting())
	{
		it = m_printing_it;
	}
	else
	{
		it = pDoc->m_symbols.begin();
	}

	int current_y = 0;

	// The font for naming with
	CFont theFont;
	theFont.CreateFont(m_view_transform.intScale( 20 ),0,0,0,400,FALSE,FALSE,FALSE,
		ANSI_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS | CLIP_LH_ANGLES,
		DEFAULT_QUALITY,DEFAULT_PITCH | FF_MODERN,"Courier New");

	CFont *oldFont = pDC->SelectObject( &theFont );

	TransformSnap s;
	s.SetGridSnap( FALSE );
	s.SetAccurateGrid(FineGrid);
	CTinyCadDoc doc;
	
	const int inter_spacing = 20;
	const int text_spacing = 25;
	int current_index = 0;
	current_y = 0;

	int height = m_view_transform.intDeScale( pDC->GetDeviceCaps( VERTRES ) );

	while (it != pDoc->m_symbols.end())
	{
		Context q( pDC, m_view_transform );

		// Display the name of this symbol
		pDC->SelectObject( &theFont );
			
		// Now access the symbol and draw it next to the name
		CPoint p;
		CDesignFileSymbol *pSymbol = (it->second).GetDesignSymbol( &doc );

		drawingCollection method;
		if (pSymbol)
		{
			pSymbol->GetMethod( p, method );
		}
		else
		{
			p = CPoint(-75,-15);
		}
		
		if ( pDC->IsPrinting() && current_y + -p.x + inter_spacing*2 > height)
		{
			delete pSymbol;
			break;
		}


		// Fill in the background rectangle
		q.theTransform.SetOrigin( CPoint( 0, -current_y) );
		if (!pDC->IsPrinting())
		{
			if (current_index == pDoc->getSelected())
			{
				q.SelectBrush( cOFFPAGE );
			}
			else
			{
				q.SelectBrush( cWHITE );
			}
			q.SelectPen( cBLACK,0,0);
			q.Rectangle( CRect(0,0,1024,-p.x+inter_spacing*2 ));
		}

		// Draw the name
		q.SetTextAlign(TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);
		q.SetBkMode( TRANSPARENT );
		q.TextOut( 5, 20, (it->second).name );
		q.TextOut( 5, 40, (it->second).description );

		// Now display the symbol
		q.theTransform.SetOrigin( CPoint( -500, -current_y) );
		CPoint old = q.theTransform.SetTRM( CPoint(0,-p.x+inter_spacing), CPoint(0,0), 2 );

		if (pSymbol)
		{
			drawingIterator paint_it = method.begin();
			while (paint_it != method.end())
			{
				(*paint_it)->Paint( q, draw_normal );
				++ paint_it;
			}

			delete pSymbol;
		}

		q.theTransform.EndTRM( old );

		// Move on to the next symbol
		it ++;

		// Note: we use p.x because we have rotated the 
		// symbol
		current_y += -p.x + inter_spacing*2;
		current_index ++;
	}

	pDC->SelectObject( oldFont );


	if (pDC->IsPrinting())
	{
		m_printing_it = it;
	}

}

/////////////////////////////////////////////////////////////////////////////
// CLibraryView diagnostics

#ifdef _DEBUG
void CLibraryView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CLibraryView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}

CLibraryDoc* CLibraryView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CLibraryDoc)));
	return (CLibraryDoc*)m_pDocument;
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CLibraryView message handlers

void CLibraryView::OnInitialUpdate() 
{
	CScrollView::OnInitialUpdate();
	
	SetScrollSizes( MM_TEXT, GetDocument()->GetDocSize() );
	
}
void CLibraryView::OnSymbolNewsymbol() 
{
	GetDocument()->EditSymbol( -1 );		
}

void CLibraryView::OnSymbolDeletesymbol() 
{
	GetDocument()->DeleteSymbol( GetDocument()->getSelected() );
	SetScrollSizes( MM_TEXT, GetDocument()->GetDocSize() );	
	Invalidate();
}

void CLibraryView::OnSymbolEditsymbol() 
{
	GetDocument()->EditSymbol( GetDocument()->getSelected() );
	
}

void CLibraryView::OnSymbolRenamesymbol() 
{
	// Determine the old name
	CLibraryFile::symbolCollection::iterator it = GetDocument()->m_symbols.begin();
	for (int i =0; it != GetDocument()->m_symbols.end() && i < GetDocument()->getSelected(); i++ )
	{
		it ++;
	}

	// If we have a selected item, then rename it
	if (it != GetDocument()->m_symbols.end())
	{
		CString name = (it->second).name;
		CDlgSymbolRename dialog(&name);
		if (dialog.DoModal() == IDOK)
		{
			GetDocument()->RenameSymbol( GetDocument()->getSelected(), name );
			Invalidate();
		}
	}
	
	
}


void CLibraryView::OnUpdateSymbolDeletesymbol(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable( GetDocument()->getSelected() != -1 );
}

void CLibraryView::OnUpdateSymbolEditsymbol(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable( GetDocument()->getSelected() != -1 );
	
}


void CLibraryView::OnUpdateSymbolRenamesymbol(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable( GetDocument()->getSelected() != -1 );
}



/////////////////////////////////////////////////////////////////////////////
// CLibraryView printing

BOOL CLibraryView::OnPreparePrinting(CPrintInfo* pInfo) 
{


	
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CLibraryView::OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo) 
{
	// Count the symbols and determine the page count and splits
	CLibraryDoc * pDoc = GetDocument();
	CLibraryFile::symbolCollection::iterator it = pDoc->m_symbols.begin();


	TransformSnap s;
	s.SetGridSnap( FALSE );
	CTinyCadDoc doc;
	m_view_transform.SetZoom( 500 );

	const int inter_spacing = 20;
	const int text_spacing = 25;
	int current_index = 0;
	int current_y = 0;

	int height = m_view_transform.intDeScale( pDC->GetDeviceCaps( VERTRES ) );
	int pages = 1;

	while (it != pDoc->m_symbols.end())
	{
		// Now access the symbol and draw it next to the name
		CPoint p;
		CDesignFileSymbol *pSymbol = (it->second).GetDesignSymbol( &doc );
		drawingCollection method;
		pSymbol->GetMethod( p, method );
		delete pSymbol;

		// Is it time to move onto the next page?
		if ( current_y + -p.x + inter_spacing*2 > height)
		{
			pages ++;
			current_y = 0;
		}

		// Move on to the next symbol
		it ++;

		// Note: we use p.x because we have rotated the 
		// symbol
		current_y += -p.x + inter_spacing*2;
		current_index ++;
	
	}

	pInfo->SetMinPage( 1 );
	pInfo->SetMaxPage( pages );

	m_printing_it = pDoc->m_symbols.begin();
	
	CScrollView::OnBeginPrinting(pDC, pInfo);
}

void CLibraryView::OnPrepareDC(CDC* pDC, CPrintInfo* pInfo) 
{
	if (pInfo)
	{

		// Count the symbols and determine the page count and splits
		CLibraryDoc * pDoc = GetDocument();
		m_view_transform.SetZoom( 500 );
		TransformSnap s;
		s.SetGridSnap( FALSE );
		CTinyCadDoc doc;

		const int inter_spacing = 20;
		const int text_spacing = 25;
		int current_index = 0;
		int current_y = 0;

		int height = m_view_transform.intDeScale( pDC->GetDeviceCaps( VERTRES ) );
		int pages = 1;

		m_printing_it = pDoc->m_symbols.begin();

		if (pInfo->m_nCurPage > 1)
		{
			while (m_printing_it != pDoc->m_symbols.end())
			{
				// Now access the symbol and draw it next to the name
				CPoint p;
				CDesignFileSymbol *pSymbol = (m_printing_it->second).GetDesignSymbol( &doc );
				drawingCollection method;
				pSymbol->GetMethod(p,method);
				delete pSymbol;

				// Is it time to move onto the next page?
				if ( current_y + -p.x + inter_spacing*2 > height)
				{
					pages ++;
					current_y = 0;

					if (pages == pInfo->m_nCurPage)
					{
						break;
					}
				}

				// Move on to the next symbol
				m_printing_it ++;

				// Note: we use p.x because we have rotated the 
				// symbol
				current_y += -p.x + inter_spacing*2;
				current_index ++;
			}
		}
	}
	else
	{
		m_view_transform.SetZoom( 100 );
	}


	CScrollView::OnPrepareDC(pDC, pInfo);
}

void CLibraryView::OnEndPrinting(CDC* pDC, CPrintInfo* pInfo) 
{
	// TODO: Add your specialized code here and/or call the base class	
	CScrollView::OnEndPrinting(pDC, pInfo);
}



/////////////////////////////////////////////////////////////////////////////
// The user is selecting a new object

void CLibraryView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// Determine the current scroll limits
	int y_hit = GetScrollPosition().y + point.y;

	// Now determine which object this lie within
	int current_index = 0;
	std::set<int>::iterator it = GetDocument()->m_heights.begin();
	int last_y_height;
	while (it != GetDocument()->m_heights.end())
	{
		// Is this a hit?
		if (*it > y_hit)
			break;

		// Move on
		current_index ++;
		last_y_height = *it;
		it ++;
	}

	GetDocument()->setSelected( current_index );
	Invalidate();
	
	CScrollView::OnLButtonDown(nFlags, point);
}


void CLibraryView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	OnSymbolEditsymbol();
	
	CScrollView::OnLButtonDblClk(nFlags, point);
}

