/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "stdafx.h"
#include "TinyCad.h"
#include "TinyCadView.h"
#include "resource.h"
#include "revision.h"
#include "colour.h"
#include "option.h"
#include "registry.h"
#include "layer.h"

#include <list>


////// The resource methods //////

// The constructor
CDocResource::CDocResource() 
{ 
	ResourceNumber = 0; 
	MergeNumber = 0; 
	InUse = FALSE; 
	next = NULL; 
}

// The virtual destructor
CDocResource::~CDocResource() 
{ 
	// Delete down the chain...
	delete next;
}


hRESOURCE CDocResource::ConvertOld(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer!=NULL && n!=pointer->MergeNumber) {
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font
	if (pointer == NULL)
		return 0;

	return pointer->ResourceNumber;
}


// Add a new resource to this list
hRESOURCE CDocResource::Add(CDocResource *NewResource,hRESOURCE NewMerge)
{
	CDocResource *pointer = this, *last = NULL;
	hRESOURCE nextNumber = 0;

	while (pointer!=NULL && !(pointer->MergeNumber == -1 && pointer->Compare(NewResource)) ) 
	{
		// Find the next index number available
		if (pointer->ResourceNumber >= nextNumber)
			nextNumber = pointer->ResourceNumber + 1;

		last = pointer;
		pointer=pointer->next;
	}

	// This resouce is already defined
	if (pointer != NULL) {
		pointer->MergeNumber = NewMerge;
		// The new resource is not required and can be deleted!
		delete NewResource;
		// Now return the resource number
		return pointer->ResourceNumber;
	}

	ASSERT(last != NULL);

	// Not defined so add the resource to the end of the list
	NewResource->ResourceNumber = nextNumber;
	NewResource->MergeNumber = NewMerge;
	last->next = NewResource;

	return nextNumber;
}


// Get a resource from the list
CDocResource *CDocResource::Get(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		if (pointer->ResourceNumber == n)
			break;
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font (the head of the list)
	if (pointer == NULL)
		pointer = this;

	return pointer;

}

// Tag a resource given a resource number
void CDocResource::Tag(hRESOURCE n)
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		if (pointer->ResourceNumber == n)
			break;
		pointer = pointer->next;
	}

	// If the number was out of range then return the default font
	if (pointer != NULL)
		pointer->InUse = TRUE;
}


void CDocResource::UnTag()
{
	CDocResource *pointer = this;

	while (pointer != NULL) {
		pointer->InUse = FALSE;
		pointer = pointer->next;
	}

}


void CDocResource::Save(CTinyCadDoc *pDesign, CArchive &theFile)
{
	// Count the number of fonts in operation
	hRESOURCE NumberOfResources = 0;
	CDocResource  *pointer = this;

	std::list<CDocResource *> resources;
	while (pointer!=NULL) 
	{
		if (pointer->InUse)
		{
			resources.push_back( pointer );
			NumberOfResources++;
		}
		pointer = pointer->next;
	}

	// Write out the number of resources
	theFile << NumberOfResources;

	// Now actually save the resources
	std::list<CDocResource *>::iterator it = resources.begin();

	while (it != resources.end()) 
	{
		pointer = *it;
		theFile << pointer->ResourceNumber;
		pointer->SaveItem(pDesign,theFile);

		it ++;
	}
}

void CDocResource::Load(CTinyCadDoc *pDesign, CArchive &theFile)
{
	// All the merge numbers to invalid
	CDocResource *pointer = this;
	while (pointer!=NULL) 
	{
		pointer->MergeNumber = -1;
		pointer = pointer->next;
	}

	// Read in the number of Fonts in this list
	WORD NumberOfResources;
	theFile >> NumberOfResources;

	hRESOURCE OldResourceNumber = 0;

	while (NumberOfResources > 0) 
	{
		theFile >> OldResourceNumber;
		LoadItem(pDesign, theFile,OldResourceNumber);
		NumberOfResources--;
	}
}






////// The Line Style Options /////


// The resource for line styles
BOOL ListOfStyles::Compare(CDocResource *o)
{
	#define  pStyle (((ListOfStyles *)o)->theLineStyle)

	return (pStyle.Style     == theLineStyle.Style
		 && pStyle.Thickness == theLineStyle.Thickness
		 && pStyle.Colour    == theLineStyle.Colour );
}


void ListOfStyles::SaveItem(CTinyCadDoc *pDesign, CArchive &theFile)
{
	theFile.Write(&theLineStyle,sizeof(LineStyle));
}


void ListOfStyles::LoadItem(CTinyCadDoc *pDesign, CArchive &theFile, hRESOURCE n)
{
	LineStyle ls;
	theFile.Read(&ls,sizeof(LineStyle));

	ListOfStyles *nFS = new ListOfStyles(ls);

	Add(nFS,n);
}




////// The Line Style Options /////


// The resource for fill styles
BOOL ListOfFillStyles::Compare(CDocResource *o)
{
	#define  pFillStyle (((ListOfFillStyles *)o)->theFillStyle)

	return pFillStyle.Colour    == theFillStyle.Colour
		&& pFillStyle.Index		== theFillStyle.Index;
}


void ListOfFillStyles::SaveItem(CTinyCadDoc *pDesign, CArchive &theFile)
{
	theFile.Write(&theFillStyle,sizeof(theFillStyle));
}


void ListOfFillStyles::LoadItem(CTinyCadDoc *pDesign, CArchive &theFile, hRESOURCE n)
{
	FillStyle ls;
	theFile.Read(&ls,sizeof(theFillStyle));

	ListOfFillStyles *nFS = new ListOfFillStyles(ls);

	Add(nFS,n);
}




////// The MetaFile Resource /////


// The resource for line styles
BOOL ListOfMetaFiles::Compare(CDocResource *o)
{
	#define  pMetaFile (((ListOfMetaFiles *)o)->theMetaFile)

	return theMetaFile == pMetaFile;
}


void ListOfMetaFiles::SaveItem(CTinyCadDoc *pDesign, CArchive &theFile)
{
  // Put the meta-file into a memory buffer
  UINT size = GetEnhMetaFileBits( theMetaFile, 0, NULL );
  BYTE *buffer = new BYTE[ size ];
  GetEnhMetaFileBits( theMetaFile, size, buffer );

  // Write the size to the archive
  theFile << size;

  // Write this data to the archive
  theFile.Write( buffer, size );

  // Now discard the buffer
  delete[] buffer;

}


void ListOfMetaFiles::LoadItem(CTinyCadDoc *pDesign, CArchive &theFile, hRESOURCE n)
{
  // Now read in the metafile...
  UINT size;
  theFile >> size;
  BYTE *buffer = new BYTE[ size ];

  // Get the metafile from the buffer
  theFile.Read( buffer, size );

  // Now create the new metafile...
  HENHMETAFILE data = SetEnhMetaFileBits( size, buffer );

  // ... and delete the buffer
  delete[] buffer;

  ListOfMetaFiles *nFS = new ListOfMetaFiles(data);

  Add(nFS,n);
}





////// The Font choosing Options //////


ListOfFonts::ListOfFonts(LOGFONT f)
{ 
	FontStyle = f; 
}


// The virtual constructs for the ListOfFonts resource
BOOL ListOfFonts::Compare(CDocResource *o)
{
	#define  pFont ((ListOfFonts *)o)->FontStyle

	return pFont.lfHeight    == FontStyle.lfHeight
		&& pFont.lfWidth     == FontStyle.lfWidth
  		&& pFont.lfWeight    == FontStyle.lfWeight
  		&& pFont.lfItalic    == FontStyle.lfItalic
  		&& pFont.lfUnderline == FontStyle.lfUnderline
  		&& pFont.lfStrikeOut == FontStyle.lfStrikeOut
  		&& pFont.lfCharSet   == FontStyle.lfCharSet
  		&& strcmp(pFont.lfFaceName,FontStyle.lfFaceName)==0;
}


void ListOfFonts::SaveItem(CTinyCadDoc *pDesign, CArchive &theFile)
{
	theFile.Write(&FontStyle,sizeof(LOGFONT));
}


void ListOfFonts::LoadItem(CTinyCadDoc *pDesign, CArchive &theFile, hRESOURCE n)
{
	LOGFONT lf;

	theFile.Read(&lf,sizeof(LOGFONT));

	ListOfFonts *nFS = new ListOfFonts(lf);

	Add(nFS,n);
}






////// The Symbol Resource /////

ListOfSymbols::~ListOfSymbols()
{
	delete theSymbol;
}

// The resource for line styles
BOOL ListOfSymbols::Compare(CDocResource *o)
{
	#define  pSymbol (((ListOfSymbols *)o)->theSymbol)
	return theSymbol->name == pSymbol->name;
}


void ListOfSymbols::SaveItem(CTinyCadDoc *pDesign, CArchive &theFile)
{
	// Save the symbol with the document...
	theSymbol->Save( theFile );
}


void ListOfSymbols::LoadItem(CTinyCadDoc *pDesign, CArchive &theFile, hRESOURCE n)
{
	// Save the symbol with the document...
	drawingCollection method;
	CDesignFileSymbol *psymbol = new CDesignFileSymbol( pDesign, method, CPoint(0,0) );
	psymbol->Load( pDesign, theFile );

  ListOfSymbols *nFS = new ListOfSymbols(psymbol);

  Add(nFS,n);

}

// Reload from the library
void ListOfSymbols::ResetSymbol()
{
	// Extract this symbol from the library and see what happens...
	CLibraryFileSymbol *new_symbol = static_cast<CTinyCadApp*>(AfxGetApp())->Extract( theSymbol->name );

	if (new_symbol)
	{
		CTinyCadDoc *pDesign = theSymbol->GetDesign();
		delete theSymbol;
		theSymbol = new_symbol->GetDesignSymbol( pDesign );
	}
}
