/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "TinyCad.h"

#include "TinyCadDoc.h"
#include "Revision.h"
#include "Option.h"
#include "Library.h"
#include "diag.h"
#include "colour.h"
#include "registry.h" 
#include "TinyCadView.h"
#include "LineUtils.h"
#include "JunctionUtils.h"
#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CTinyCadDoc

IMPLEMENT_DYNCREATE(CTinyCadDoc, CDocument)

BEGIN_MESSAGE_MAP(CTinyCadDoc, CDocument)
	//{{AFX_MSG_MAP(CTinyCadDoc)
		// NOTE - the ClassWizard will add and remove mapping macros here.
		//    DO NOT EDIT what you see in these blocks of generated code!
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CTinyCadDoc construction/destruction

CTinyCadDoc::CTinyCadDoc()
{
  theOptions.Init( this );

  selectable = NULL;

  MaxXSize = 297 * PIXELSPERMM;
  MaxYSize = 210 * PIXELSPERMM;

  Title = "";
  Author = "";
  Revision = "1.0";
  DocNumber = "";
  Organisation = "";
  Sheets = "1 of 1";
  ShowDetails = TRUE;
  DateStr = CTime::GetCurrentTime().Format("%d %B %Y");
  ShowRulers = FALSE;

  Undo_Level = 0;
  Top_Level = 0;
  Bottom_Level = 0;
  Undo_Inc = FALSE;

  edit=NULL;
  NextRepeat = NULL;
  LastRepeat = NULL;
  RepeatDir = CPoint(0,1);
  NameDir = 1;
  PinDir = 1;


  part = 0;


  // Now select a drawing object
  SelectObject(new CDrawEditItem(this));

	// Attempt to load the defaults for the grid
	WORD hold;
	Registry theRegistry;

	m_snap.SetAccurateGrid( theRegistry.Get("GridSpacing",&hold) ? hold : NormalGrid );
	m_snap.SetGridSnap( theRegistry.Get("GridSnap",&hold) ? hold : TRUE );

	ShowRulers = theRegistry.Get("ShowDesignRulers",&hold) ? hold : FALSE;
}


CTinyCadDoc::~CTinyCadDoc()
{
	Registry theRegistry;
	theRegistry.Place("GridSpacing",m_snap.GetAccurateGrid());
	theRegistry.Place("GridSnap",m_snap.GetGridSnap());

	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;
		delete pointer;
		++ it;
	}
	m_drawing.erase( m_drawing.begin(), m_drawing.end() );

	UnSelect();

	// Remove the undo/redo list
	it = m_redo.begin();
	while (it != m_redo.end()) 
	{
		delete *it;
		++ it;
	}
	m_redo.erase( m_redo.begin(), m_redo.end() );

	// Delete the editing tool
	delete edit;
	delete NextRepeat;
	delete LastRepeat;
}

BOOL CTinyCadDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

  MaxXSize = 297 * PIXELSPERMM;
  MaxYSize = 210 * PIXELSPERMM;


  // Has the page size been stored in the registry?
  Registry theRegistry;
  int Fd;
  CString GetString;
  if (theRegistry.Get("PageSize",GetString) && (Fd=GetString.Find(','))!=-1 ) {
	CPoint NewSize;
	NewSize.y = atoi(GetString.Mid(Fd+1));
	NewSize.x = atoi(GetString.Left(Fd));
	SetPageBoundries(NewSize);		
  }


  SetPart(0);

  SelectObject(new CDrawEditItem(this));
  UnSelect();

	// Remove all objects from memory
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;
		delete pointer;
		++ it;
	}
	m_drawing.erase( m_drawing.begin(), m_drawing.end() );


  // Now delete the Redo List
	it = m_redo.begin();
	while (it != m_redo.end()) 
	{
		delete *it;
		++ it;
	}

	m_redo.erase( m_redo.begin(), m_redo.end() );

  Undo_Level = 0;
  Top_Level = 0;
  Bottom_Level = 0;
  Undo_Inc = FALSE;

  // reset the design
  Title = "";
  Author = "";
  Revision = "1.0";
  DocNumber = "";
  Organisation = "";
  Sheets = "1 of 1";

  RepeatDir = CPoint(0,1);
  NameDir = 1;
  PinDir = 1;

  part = 0;

  return TRUE;
}


// Iterate through the drawing
drawingIterator CTinyCadDoc::GetDrawingBegin()
{
	return m_drawing.begin();
}

drawingIterator CTinyCadDoc::GetDrawingEnd()
{
	return m_drawing.end();
}



/////////////////////////////////////////////////////////////////////////////
// CTinyCadDoc serialization

void CTinyCadDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		theOptions.UnTag();
		Save( ar, TRUE, FALSE );
	}
	else
	{
		drawingCollection drawing;

		if (ReadFile(ar, TRUE, drawing) )
		{
			m_drawing.insert( m_drawing.end(), drawing.begin(), drawing.end() );
			UnSelect();
			SetModifiedFlag( FALSE );
		}
	}
}

BOOL CTinyCadDoc::Import( CArchive& ar )
{
	drawingCollection drawing;

	if (ReadFile(ar, TRUE, drawing) )
	{
		m_drawing.insert( m_drawing.end(), drawing.begin(), drawing.end() );
		UnSelect();
		SetModifiedFlag( FALSE );

		// Now select the objects
		UnSelect();
		drawingIterator it = drawing.begin();
		while (it != drawing.end())
		{
			Select( *it );
			++ it;
		}

		return TRUE;
	}

	return FALSE;
}


/////////////////////////////////////////////////////////////////////////////
// CTinyCadDoc diagnostics

#ifdef _DEBUG
void CTinyCadDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CTinyCadDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CTinyCadDoc commands




// Convert this document into a metafile for pasting into
// other applications
HENHMETAFILE CTinyCadDoc::CreateMetafile( CDC &ref_dc )
{

	// Calculate the boundries
	CRect rect = CRect(0,0,0,0);

	if (IsSelected())
	{
		BOOL first = TRUE;

		selectIterator it = GetSelectBegin();
		while ( it != GetSelectEnd() ) 
		{
			CDrawingObject *pointer=*it;

			if (first) 
			{
				rect.top = pointer->a.y;
				rect.bottom = pointer->a.y;
				rect.left = pointer->a.x;
				rect.right = pointer->a.x;
				first = FALSE;
			}

			rect.top 	 = min(rect.top,min(pointer->a.y,pointer->b.y));
			rect.bottom = max(rect.bottom,max(pointer->a.y,pointer->b.y));
			rect.left 	 = min(rect.left,min(pointer->a.x,pointer->b.x));
			rect.right  = max(rect.right,max(pointer->a.x,pointer->b.x));

			++ it;
		}
	}
	else
	{
		rect = CRect( 0,0, MaxXSize, MaxYSize );
	}

	rect.InflateRect( 0,0, 10,10 );


	// Transform the origin so the design lies at 0,0
	Transform newTransform;


	int iWidthMM = ref_dc.GetDeviceCaps(HORZSIZE); 
	int iHeightMM = ref_dc.GetDeviceCaps(VERTSIZE); 
	int iWidthPels = ref_dc.GetDeviceCaps(HORZRES); 
	int iHeightPels = ref_dc.GetDeviceCaps(VERTRES); 
 
	// Convert client coordinates to .01-mm units. 
	// Use iWidthMM, iWidthPels, iHeightMM, and 
	// iHeightPels to determine the number of 
	// .01-millimeter units per pixel in the x- 
	//  and y-directions. 
 
	rect.left = (rect.left * iWidthMM * 100)/iWidthPels; 
	rect.top = (rect.top * iHeightMM * 100)/iHeightPels; 
	rect.right = (rect.right * iWidthMM * 100)/iWidthPels; 
	rect.bottom = (rect.bottom * iHeightMM * 100)/iHeightPels;


	// Create the metafile device context. 
	HDC hdcMeta = CreateEnhMetaFile(ref_dc.m_hDC,NULL, &rect, Title); 
 
	if (!hdcMeta) 
	{
		AfxMessageBox("Cannot create enhanced metafile for the copy");
		return NULL;
	}

	CDC theMetaDC;
	theMetaDC.Attach( hdcMeta );

	// Now render into this dc
	newTransform.SetZoom( 100 );

	// Create the DC and render...
	{
		Context dc(&theMetaDC,newTransform);

		if (IsSelected())
		{
			// Draw just the selected part
			selectIterator it = GetSelectBegin();
			while ( it != GetSelectEnd() ) 
			{
				(*it)->Paint(dc,draw_normal);
				++ it;
			}
		}
		else
		{
			// Draw the whole design into the metafile
			drawingIterator it = GetDrawingBegin();
			while (it != GetDrawingEnd()) 
			{
				(*it)->Paint(dc,draw_normal);
				++ it;
			}

			// Show the design details
			DisplayDetails( dc );

			// Draw the design rulers
			DisplayRulers(dc);
		}

	}

	theMetaDC.Detach();

	return CloseEnhMetaFile( hdcMeta );
}


// Update the page boundries etc, using a printer device context
void CTinyCadDoc::SetPageBoundries(PRINTDLG &pd)
{
  CDC *dc = CDC::FromHandle( pd.hDC );

  // Change the scaling to print correct size
  MaxXSize=dc->GetDeviceCaps(HORZSIZE)*PIXELSPERMM;
  MaxYSize=dc->GetDeviceCaps(VERTSIZE)*PIXELSPERMM;
}


// Make all symbol aware of the fact the libraries have changed
void CTinyCadDoc::ResetSymbols()
{
	theOptions.ResetSymbols();
}


// Tag all the resources being used by this design
void CTinyCadDoc::TagAllResources()
{
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		(*it)->TagResources();
		++ it;
	}
}

// Remove all errors from this design
void CTinyCadDoc::DeleteErrors()
{
  // Get rid of any drawing tool
  SelectObject(new CDrawEditItem(this));

	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		drawingIterator current = it;
		++ it;
		
		CDrawingObject *pointer = *current;
		if (pointer->GetType()==xError) 
		{
			m_drawing.erase(current);
			delete pointer;
		}
	}

	Invalidate();
}


// Is this document editing a library?
bool CTinyCadDoc::IsLibInUse( CLibraryFile *lib )
{
	return false;
}

BOOL CTinyCadDoc::IsSelected()
{ 
	return m_selected.size() > 0;
}

selectIterator CTinyCadDoc::GetSelectBegin()
{
	return m_selected.begin();
}

selectIterator CTinyCadDoc::GetSelectEnd()
{
	return m_selected.end();
}

BOOL CTinyCadDoc::IsSelected( CDrawingObject * p )
{
	return m_selected.find( p ) != m_selected.end();
}
BOOL CTinyCadDoc::IsSingleItemSelected()
{
	return m_selected.size() == 1; 
}

CDrawingObject* CTinyCadDoc::GetSingleSelectedItem()
{
	if (m_selected.size() == 1)
	{
		return *(m_selected.begin());
	}
	else
	{
		return NULL;
	}
}




// Add a new object to this drawing
void CTinyCadDoc::Add(CDrawingObject *NewObject)
{
  // Is this a valid object?
  if (NewObject == NULL)
	return;

  // Do we need to increment the Undo Level?
  FlushRedo();

  // Set the Undo Level of the new object
  NewObject->Undo_Level = Top_Level;
  NewObject->Delete_Level = -1;

  // Store the object in the linked list
  m_drawing.insert( m_drawing.end(), NewObject );

  // Set the dirty flag
  if (NewObject->GetType()!=xError)
  {
	SetModifiedFlag( TRUE );
  }
}


void CTinyCadDoc::Erase( drawingIterator it )
{
	(*it)->Display();
	m_drawing.erase( it );
	SetModifiedFlag( TRUE );
}


// Undo the last action
void CTinyCadDoc::Undo()
{
  Undo_Level --;
  SwapLists();
  Invalidate();
}

// Redo the last action
void CTinyCadDoc::Redo()
{
  if (Undo_Level<Top_Level) {
	Undo_Level ++;
	SwapLists();
	Invalidate();
  }
}


// Swap non-displayable items into Redo List and displable items from redo
// list into normal list
void CTinyCadDoc::SwapLists()
{
  CJunctionUtils j(this);

  // Move the objects from the Redo List to the object list
  drawingIterator redo_it = m_redo.begin();
  while (redo_it != m_redo.end()) 
  {
	  drawingIterator current = redo_it ;
	  ++ redo_it;
	  CDrawingObject *pointer = *current;

	if (pointer->DoesExist()) 
	{
		m_redo.erase( current );
		j.AddObjectToTodo( pointer );
		m_drawing.insert( m_drawing.begin(), pointer );
	}
  }


  // Move the no longer valid objects onto the Redo List
  	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		drawingIterator current = it;
		++ it;

		CDrawingObject *pointer = *current;

		if ( !(pointer->DoesExist()) ) 
		{
			m_drawing.erase( current );

			j.AddObjectToTodo( pointer );
			m_redo.insert( m_redo.begin(), pointer );
		}

	}

	j.CheckTodoList();
}


// Flush the Redo Buffer
void CTinyCadDoc::FlushRedo()
{
  // Delete all objects that MAX_UNDO's ago won't exist!
  int Old_Undo_Level = Undo_Level;
  Undo_Level -= MAX_UNDO;

  // Does the Redo buffer need flushing?
  if (Top_Level!=Undo_Level) 
  {
	// Delete all objects from the Redo list
	drawingIterator it = m_redo.begin();
	while (it!=m_redo.end()) 
	{
		drawingIterator current = it;
		++ it;
		CDrawingObject *pointer = *current;

		if ( pointer->Delete_Level == -1 || pointer->Delete_Level<Undo_Level ) 
		{
			m_redo.erase( current );
			delete pointer;
		} 
	}
  }
  Bottom_Level = max(Bottom_Level,Undo_Level);
  Undo_Level = Old_Undo_Level;

  // Do we need to increment the Undo Level?
  if (Undo_Inc) 
  {
	Undo_Level++;
	Top_Level = Undo_Level;
	Undo_Inc = FALSE;
  }


	// Remove any deletes in the design list
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		CDrawingObject *pointer = *it;
		if (pointer->Delete_Level != -1)
		{
			pointer->Delete_Level = -1;
		}
		++ it;
	}
}




// The object selection functions
// This selects objects in a box
// (We don't select construction objects)
void CTinyCadDoc::Select(CPoint p1,CPoint p2)
{
  int left=Min(p1.x,p2.x);
  int right=Max(p1.x,p2.x);
  int top=Min(p1.y,p2.y);
  int bottom=Max(p1.y,p2.y);

  UnSelect();

	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		CDrawingObject *obj = *it;

		if (    obj->IsInside(left,right,top,bottom) 
			&& !obj->IsConstruction()
			&& 
			(obj->GetType() != xJunction || !GetOptions()->GetAutoJunc() ) )
		{
			obj->Display();
			Select( obj );
		} 
  		
		++ it;
	}
}


// Select an object
void CTinyCadDoc::Select(CDrawingObject *obj )
{
	m_selected.insert( obj );
}

// Unselect a single object
void CTinyCadDoc::UnSelect( CDrawingObject *p )
{
	selectIterator it = m_selected.find( p );
	if (it != m_selected.end())
	{
		m_selected.erase( it );
	}
}


// Unselect the objects
void CTinyCadDoc::UnSelect()
{
	// Get rid of any currently selected objects
	selectIterator it = GetSelectBegin();
	while ( it != GetSelectEnd() ) 
	{
		(*it)->Display();
		++ it;
	}

	m_selected.erase( m_selected.begin(), m_selected.end() );
}


void CTinyCadDoc::DupForUndo( CDrawingObject *pObject )
{
	CDrawingObject *pNewObject = pObject->Store();
	m_drawing.erase( -- m_drawing.end() );

	pNewObject->Delete_Level = Undo_Level;
	pNewObject->Undo_Level = pObject->Undo_Level;

	m_redo.insert( m_redo.begin(), pNewObject );

	pObject->Undo_Level = Top_Level;

    SetModifiedFlag( TRUE );
}


// Change the Z-order of the selected objects
void CTinyCadDoc::BringToFront()
{
	if (!IsSelected())
		return;

	FlushRedo();

	drawingCollection selectedObjects;
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		drawingIterator current = it;
		++ it;

		CDrawingObject *pointer = *current;
		if (IsSelected( pointer ))
		{
			m_drawing.erase( current );
			selectedObjects.push_back( pointer );
		}
	}


	// Now push them back on (in the right order)
	drawingCollection::iterator itx = selectedObjects.begin();
	while (itx != selectedObjects.end())
	{
		m_drawing.push_back( *itx );
		(*itx)->Display();
		++ itx;
	}

}

// Change the Z-order of the selected objects
void CTinyCadDoc::SendToBack()
{
	if (!IsSelected())
		return;

	FlushRedo();

	drawingCollection selectedObjects;
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		drawingIterator current = it;
		++ it;

		CDrawingObject *pointer = *current;
		if (IsSelected( pointer ))
		{
			m_drawing.erase( current );
			selectedObjects.push_back( pointer );
		}
	}

	// Now push them back on (in the right order)
	drawingCollection::reverse_iterator itx = selectedObjects.rbegin();
	while (itx != selectedObjects.rend())
	{
		m_drawing.push_front( *itx );
		(*itx)->Display();
		++ itx;
	}
}

// Delete the selected objects!
void CTinyCadDoc::SelectDelete()
{
  if (!IsSelected())
	return;

  FlushRedo();

	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		drawingIterator current = it;
		++ it;

		CDrawingObject *pointer = *current;
		if (IsSelected( pointer ))
		{
			if (pointer == GetSelectable())
			{
				SetSelectable( NULL );
			}

			m_drawing.erase( current );
			m_redo.push_front( pointer );
		}
	}


  // Unselect all objects
  UnSelect();

  SetModifiedFlag( TRUE );
}




// Place the Selected objects into the Undo List in the original position
void CTinyCadDoc::SelectUndoMove(CPoint r)
{

  if (!IsSelected())
	return;

  	selectIterator it = GetSelectBegin();
	while ( it != GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;

		// Now move the object
		obj->a.x += r.x;
		obj->a.y += r.y;
		obj->b.x += r.x;
		obj->b.y += r.y;

		// Make a copy of the selected objects (so as to perform add/delete operation for Redo List)
		DupForUndo( obj );

		// Now move the object back...
		obj->a.x -= r.x;
		obj->a.y -= r.y;
		obj->b.x -= r.x;
		obj->b.y -= r.y;

		++ it;
	}


  SetModifiedFlag( TRUE );

}





// Save the current design
BOOL CTinyCadDoc::Save(BOOL GetName,BOOL SaveSelect)
{

  CString theFileName = GetPathName();

  // Get the filename if necessary
  if (GetName) { 
        OPENFILENAME ofn;
        char szFile[256], szFileTitle[256];

       	strcpy(szFile, "*.dsn\0");

        ofn.lStructSize = sizeof(OPENFILENAME);
        ofn.hwndOwner = AfxGetMainWnd()->m_hWnd;
	  	ofn.lpstrFilter = "Designs (*.dsn)\0*.dsn\0All files (*.*)\0*.*\0\0";
        ofn.lpstrCustomFilter = NULL;
        ofn.nMaxCustFilter = 0;
        ofn.nFilterIndex = 1;
        ofn.lpstrFile = szFile;
        ofn.nMaxFile = sizeof(szFile);
        ofn.lpstrFileTitle = szFileTitle;
        ofn.nMaxFileTitle = sizeof(szFileTitle);
        ofn.lpstrInitialDir = NULL;
        ofn.lpstrTitle = NULL;
        ofn.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
        ofn.nFileOffset = 0;
        ofn.nFileExtension = 0;
        ofn.lpstrDefExt = "dsn";

        if (!GetSaveFileName(&ofn))
        	return FALSE;

		theFileName = szFile;
  }


  BOOL r;

  // Show the busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );


  // Is this file to be saved as text or as a design file?
  	CFile theFile;

  	// Open the file for saving as a CFile for a CArchive
  	r =  theFile.Open(theFileName, CFile::modeCreate | CFile::modeWrite);

	if (r) {
  		// Now save the file
		CArchive ar( &theFile, CArchive::store );
  		Save(ar,TRUE,SaveSelect);

  		SetModifiedFlag( FALSE );
  	}

  // Turn the cursor back to normal
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );

  // Was there an error opening the file?
  if (!r)
	// Could not open file to start saving
	Message(IDS_ABORTSAVE,MB_ICONEXCLAMATION);

  return r;
}





// Load a new design
BOOL CTinyCadDoc::Import()
{

  OPENFILENAME ofn;
  char szFile[256], szFileTitle[256];

  strcpy(szFile, "*.dsn\0");

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = AfxGetMainWnd()->m_hWnd;
  ofn.lpstrFilter = "Designs (*.dsn)\0*.dsn\0All files (*.*)\0*.*\0\0";
  ofn.lpstrCustomFilter = NULL;
  ofn.nMaxCustFilter = 0;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile = szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;
  ofn.lpstrTitle = NULL;
  ofn.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
  ofn.nFileOffset = 0;
  ofn.nFileExtension = 0;
  ofn.lpstrDefExt = "dsn";
  ofn.lpstrTitle = "Import";
 
  if (!GetOpenFileName(&ofn))
       	return FALSE;

  // Do we use the text loading procedure or load from a CArchive?
	CFile theFile;
	if (!theFile.Open(szFile, CFile::modeRead)) 
	{
		// Could not open file to start loading
		Message(IDS_ABORTLOAD,MB_ICONEXCLAMATION);
		return FALSE;
	}

	CArchive ar( &theFile, CArchive::load );
  	return Import( ar );
}



void CTinyCadDoc::ForceSetCursor()
{
	UpdateAllViews( NULL, DOC_UPDATE_SETCURSOR, NULL );
}


// Redraw our window
void CTinyCadDoc::Invalidate()
{
	UpdateAllViews( NULL, DOC_UPDATE_INVALIDATE, NULL );
}

void CTinyCadDoc::InvalidateRect( CRect r, BOOL erase, int grow, BOOL outline_only )
{
	if (outline_only)
	{
		CRect r1( r.left, r.top, r.right, r.top );
		CRect r2( r.right, r.top, r.right, r.bottom );
		CRect r3( r.right, r.bottom, r.left, r.bottom );
		CRect r4( r.left, r.bottom, r.left, r.top );
		InvalidateRect( r1, erase, grow );
		InvalidateRect( r2, erase, grow );
		InvalidateRect( r3, erase, grow );
		InvalidateRect( r4, erase, grow );
	}
	else
	{
		doc_invalidrect hint;
		hint.r = r;
		hint.grow = grow;
		UpdateAllViews( NULL, erase ? DOC_UPDATE_INVALIDRECTERASE : DOC_UPDATE_INVALIDRECT, &hint );
	}
}

// No obvious answer to this one, but it is
// a function that should go in here...
CSize CTinyCadDoc::GetTextExtent(const char *str, int style )
{
	Transform dummy;
	Context dc(AfxGetMainWnd(), dummy );
	dc.SelectFont(*GetOptions()->GetFont(style),0);
	return dc.GetTextExtent( str );
}


// Set the next repeat object for the editing view
void CTinyCadDoc::SetRepeat( CDrawingObject *pObject )
{
  // Get rid of the last repeat
  if (LastRepeat!=NULL) {
	delete LastRepeat;
  }

  LastRepeat = NextRepeat;

  NextRepeat = pObject;  
}

void CTinyCadDoc::Repeat()
{
  if (NextRepeat!=NULL) 
  {
	NextRepeat->Display();
	NextRepeat->Store();
  }
}

void CTinyCadDoc::SetRepeatDir( int NewDir )
{
  switch(NewDir) {
	case 0: // Up
		RepeatDir = CPoint(0,-1);
		break;
	case 1: // Down
		RepeatDir = CPoint(0,1);
		break;
	case 2: // Left
		RepeatDir = CPoint(-1,0);
		break;
	case 3: // Right
		RepeatDir = CPoint(1,0);
		break;
  }
}


// Select a new editing object
void CTinyCadDoc::SelectObject(CDrawingObject *NewO )
{ 
	if (edit)
	{
		edit->EndEdit();
		delete edit;
	}

	if (NewO)
	{
		NewO->BeginEdit(FALSE);
	}

  edit=NewO;

  // Increment the undo level
  IncUndoLevel();
}




#define BOXWIDTH	400
#define	LINEHEIGHT	18
#define RULERHEIGHT 15

// Draw the design rulers
void CTinyCadDoc::DisplayRulers(Context &dc)
{
	if (!ShowRulers)
		return;

	// Select the correct pen
	dc.SelectPen(PS_SOLID,1,cLINE);

	// Draw the outline box

	CPoint tl1 = CPoint(0, 0);
	CPoint tl2 = CPoint(RULERHEIGHT+ 2, RULERHEIGHT+ 2);
	CPoint br2 = CPoint(MaxXSize-RULERHEIGHT - 2,MaxYSize-RULERHEIGHT- 2);
	CPoint br1 = CPoint(MaxXSize - 2,MaxYSize - 2);

	// Draw the 4 lines that make up the outer boarder
	dc.MoveTo( tl1 );
	dc.LineTo( CPoint( br1.x, tl1.y ) );
	dc.LineTo( br1 );
	dc.LineTo( CPoint( tl1.x, br1.y ) );
	dc.LineTo( tl1 );

	// Draw the 4 lines that make up the inner boarder
	dc.MoveTo( tl2 );
	dc.LineTo( CPoint( br2.x, tl2.y ) );
	dc.LineTo( br2 );
	dc.LineTo( CPoint( tl2.x, br2.y ) );
	dc.LineTo( tl2 );

	// Now draw the cross bracings
	dc.MoveTo( tl1 );
	dc.LineTo( tl2 );
	dc.MoveTo( br1 );
	dc.LineTo( br2 );
	dc.MoveTo( CPoint( br1.x, tl1.y ) );
	dc.LineTo( CPoint( br2.x, tl2.y ) );
	dc.MoveTo( CPoint( tl1.x, br1.y ) );
	dc.LineTo( CPoint( tl2.x, br2.y ) );

	// Now lay out in a 5 x 7 layout...
	dc.SetTextAlign(TA_CENTER | TA_BOTTOM | TA_NOUPDATECP);
	dc.SetTextColor(cBLACK);
	dc.SetROP2(R2_COPYPEN);
	dc.SetBkMode(TRANSPARENT);


	// Select the correct font
	dc.SelectFont(*GetOptions()->GetFont(fPIN),3);

	int cols = 7;
	int rows = 5;

	if (MaxXSize < MaxYSize)
	{
		cols = 5;
		rows = 7;
	}

	int split = (br1.x - tl1.x) / cols;
	for (int col = 0; col < cols; col ++)
	{
		CString s;
		s.Format("%d",col + 1);

		dc.MoveTo( CPoint( col * split, tl1.y ));
		dc.LineTo( CPoint( col * split, tl2.y ));
		dc.TextOut( split * col + split/2, tl2.y - 2, s );

		dc.MoveTo( CPoint( col * split, br1.y ));
		dc.LineTo( CPoint( col * split, br2.y ));
		dc.TextOut( split * col + split/2, br1.y - 2, s );

	}

	// Select the correct font
	dc.SelectFont(*GetOptions()->GetFont(fPIN),0);

	split = (br1.y - tl1.y) / rows;
	for (int row = 0; row < rows; row ++)
	{
		CString s;
		s.Format("%c", 'A' + row );

		dc.MoveTo( CPoint( tl1.x, row * split));
		dc.LineTo( CPoint( tl2.x, row * split));
		dc.TextOut( tl2.x - 2, split * row + split/2, s );

		dc.MoveTo( CPoint( br1.x, row * split));
		dc.LineTo( CPoint( br2.x, row * split));
		dc.TextOut( br1.x - 1, split * row + split/2, s );
	}

}


// Draw the details box
void CTinyCadDoc::DisplayDetails(Context &dc)
{
  // Do we display the details?
  if (!ShowDetails)
	return;

  // Does the date need to be updated?
  // if (IsModified())
  //  Date = CTime::GetCurrentTime();

  // Select the correct pen
  dc.SelectPen(PS_SOLID,1,cLINE);
  // Select the correct font
  dc.SelectFont(*GetOptions()->GetFont(fTEXT),3);

  // Draw the box to house it all
  CPoint tl = CPoint(MaxXSize - BOXWIDTH-2, MaxYSize - LINEHEIGHT*9 -2);
  CPoint br = CPoint(MaxXSize-2,MaxYSize-2);

  // Move if necessary...
  if (ShowRulers)
  {
	  tl -= CPoint( RULERHEIGHT, RULERHEIGHT );
	  br -= CPoint( RULERHEIGHT, RULERHEIGHT );
  }

  int LineHeight = LINEHEIGHT;
  int TextSpace = LineHeight/2;
  int BottomRow = (br.x-tl.x)/3;
  int MiddleRow = br.x-(br.x-tl.x)/5;

  // Now draw the outline
  dc.MoveTo(tl);
  dc.LineTo(CPoint(br.x,tl.y));
  dc.LineTo(br);
  dc.LineTo(CPoint(tl.x,br.y));
  dc.LineTo(tl);

  // Draw the horizontal lines
  dc.MoveTo(CPoint(tl.x,tl.y+LineHeight*2));
  dc.LineTo(CPoint(br.x,tl.y+LineHeight*2));
  dc.MoveTo(CPoint(tl.x,tl.y+LineHeight*5));
  dc.LineTo(CPoint(br.x,tl.y+LineHeight*5));
  dc.MoveTo(CPoint(tl.x,tl.y+LineHeight*7));
  dc.LineTo(CPoint(br.x,tl.y+LineHeight*7));

  // Draw the vertical lines
  dc.MoveTo(CPoint(MiddleRow,tl.y+LineHeight*5));
  dc.LineTo(CPoint(MiddleRow,tl.y+LineHeight*7));
  dc.MoveTo(CPoint(tl.x+BottomRow,tl.y+LineHeight*7));
  dc.LineTo(CPoint(tl.x+BottomRow,tl.y+LineHeight*9));
  dc.MoveTo(CPoint(MiddleRow,tl.y+LineHeight*7));
  dc.LineTo(CPoint(MiddleRow,tl.y+LineHeight*9));

  // Add the text
  dc.SetTextAlign(TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);
  dc.SetTextColor(cBLACK);
  dc.SetROP2(R2_COPYPEN);
  dc.SetBkMode(TRANSPARENT);

  dc.TextOut(tl.x+TextSpace,tl.y+LineHeight,"Title");
  dc.TextOut(tl.x+TextSpace,tl.y+LineHeight*3,"Author");
  dc.TextOut(tl.x+TextSpace,tl.y+LineHeight*6,"File");
  dc.TextOut(tl.x+TextSpace,tl.y+LineHeight*8,"Revision");
  dc.TextOut(MiddleRow+TextSpace,tl.y+LineHeight*6,"Document");
  dc.TextOut(tl.x+BottomRow+TextSpace,tl.y+LineHeight*8,"Date");
  dc.TextOut(MiddleRow+TextSpace,tl.y+LineHeight*8,"Sheets");

  // Add the actual data!  
  dc.TextOut(tl.x+TextSpace*2,tl.y+LineHeight*2,Title);
  dc.TextOut(tl.x+TextSpace*2,tl.y+LineHeight*4,Author);
  dc.TextOut(tl.x+TextSpace*2,tl.y+LineHeight*5,Organisation);
  dc.TextOut(tl.x+TextSpace*2,tl.y+LineHeight*7,MiddleRow - tl.x - TextSpace * 4, GetPathName());
  dc.TextOut(tl.x+TextSpace*2,tl.y+LineHeight*9,Revision);
  dc.TextOut(MiddleRow+TextSpace*2,tl.y+LineHeight*7,DocNumber);
  dc.TextOut(tl.x+BottomRow+TextSpace*2,tl.y+LineHeight*9,DateStr);
  dc.TextOut(MiddleRow+TextSpace*2,tl.y+LineHeight*9,Sheets);

}


void CTinyCadDoc::SetSelectable(CDrawingObject *obj)
{
	// Only redraw if the selectable changes
	if (obj != selectable)
	{
		if (selectable)
		{
			selectable->Display( FALSE );
		}

		selectable = obj;

		if (selectable)
		{
			selectable->Display( FALSE );
		}
	}
}

CDrawingObject* CTinyCadDoc::GetSelectable()
{
	return selectable;
}

BOOL CTinyCadDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	// Is this a emf file?
	CString path = lpszPathName;
	CString extension;
	int brk = path.ReverseFind('.');
	if (brk != -1)
	{
		extension = path.Mid(brk);
	}
	if (extension.CompareNoCase(".emf") == 0)
	{
		// Load a wmf
		HENHMETAFILE data = (HENHMETAFILE)::GetEnhMetaFile( lpszPathName );
		CDrawMetaFile *pObject = new CDrawMetaFile( this );
		pObject->setMetaFile( data );
		Add( pObject );
	}
	else
	{
		// Load normally
		if (!CDocument::OnOpenDocument(lpszPathName))
			return FALSE;
	}
	
	return TRUE;
}


// Called after a paste or import to enable the
// document to sort out the imported block when
// necessary
void CTinyCadDoc::PostPaste()
{
}



CPoint CTinyCadDoc::GetStickyPoint( CPoint q, BOOL pins, BOOL wires, BOOL &is_stuck, BOOL &is_junction )
{
  CPoint r(0,0);
  bool first = true;
  double min_distance = 0;
  int items = 0;

  if (!GetOptions()->GetAutoSnap())
  {
  	  is_stuck = FALSE;
	  is_junction = FALSE;
	  return q;
  }


  // Search for methods, and look at their pins
	drawingIterator it = GetDrawingBegin();
	while (it != GetDrawingEnd()) 
	{
		CDrawingObject *ObjPtr = *it;

		int hold;
		CPoint tr;

		switch (ObjPtr->GetType()) 
		{
			case xWire:
				#define theLine ((CDrawLine*)ObjPtr)
				if (wires)
				{
					CPoint d;
					CLineUtils l( theLine->a, theLine->b );
					double distance = l.DistanceFromPoint( q, d );

					if (d == r)
					{
						items ++;
					}

					if (first || distance < min_distance)
					{
						if (r != d)
						{
							// have we split this wire?
							if (d != theLine->a && d != theLine->b)
							{
								items = 2;
							}
							else
							{
								items = 1;
							}
						}
						r = d;
						first = false;
						min_distance = distance;
					}
				}
			break;
			case xMethodEx3:
				if (pins)
				{

					
					CDrawMethod *theMethod = static_cast<CDrawMethod*>(ObjPtr);
					hold = GetPart();
					SetPart(theMethod->GetSubPart());

					drawingCollection method;
					((CDrawMethod *)ObjPtr)->ExtractSymbol(tr,method);

					drawingIterator it = method.begin();
					while (it != method.end()) 
					{
						CDrawingObject *pointer = *it;

						CDrawPin *thePin = static_cast<CDrawPin*>(pointer);
						if (pointer->GetType()==xPinEx && !(thePin->IsPower()) && !(thePin->IsInvisible()))
						{
							// This is a valid pin...
							CPoint d = thePin->GetPosition(theMethod);
							double dx = d.x - q.x;
							double dy = d.y - q.y;
							double distance = sqrt(dx*dx + dy*dy);

							if (r == d)
							{
								items ++;
							}

							if (first || distance < min_distance)
							{
								if (r != d)
								{
									items = 1;
								}

								r = d;
								first = false;
								min_distance = distance;
							}
						}

						++ it;
					}
					SetPart(hold);
				}
			break;
			case xBusSlash:
				if (pins)
				{
					CPoint d1 = ObjPtr->a;
					double d1x = d1.x - q.x;
					double d1y = d1.y - q.y;
					double distance1 = sqrt(d1x*d1x + d1y*d1y);

					CPoint d2 = ObjPtr->b;
					double d2x = d2.x - q.x;
					double d2y = d2.y - q.y;
					double distance2 = d2x*d2x + d2y*d2y;

					// Which point to use?
					if (distance2 < distance1)
					{
						distance1 = distance2;
						d1 = d2;
					}

					if (r == d1)
					{
						items ++;
					}

					if (first || distance1 < min_distance)
					{
						if (r != d1)
						{
							items = 1;
						}

						r = d1;
						first = false;
						min_distance = distance1;
					}
				}
				break;
			case xPower:
				if (pins)
				{
					// This is a valid pin...
					CPoint d = ObjPtr->a;
					double dx = d.x - q.x;
					double dy = d.y - q.y;
					double distance = sqrt(dx*dx + dy*dy);

					if (r == d)
					{
						items ++;
					}

					if (first || distance < min_distance)
					{
						if (r != d)
						{
							items = 1;
						}

						r = d;
						first = false;
						min_distance = distance;
					}
				}
			}

		++ it;
	}



  if (!first && min_distance < GetOptions()->GetAutoSnapRange())
  {
	  is_stuck = TRUE;
	  is_junction = GetOptions()->GetAutoJunc() ? items > 1 : FALSE;
	  return r;
  }
  else
  {
	  is_junction = FALSE;
	  is_stuck = FALSE;
  }

  return q;
}

// Duplicate the selected objects
void CTinyCadDoc::SelectDup()
{
	selectIterator it = GetSelectBegin();
	while ( it != GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;

		obj->Store();
		obj->Display();
		++ it;
	}
}


// Move the selected items (the CPoint is the relative shift)
void CTinyCadDoc::SelectMove(CPoint r)
{
  CJunctionUtils j( this );

  if (!IsSelected())
	return;


	// Remove all the objects from the screen
  	selectIterator it = GetSelectBegin();
	while ( it != GetSelectEnd() ) 
	{
		CDrawingObject *obj=*it;

 		// Redraw it's old position
		obj->Display();

		// Add this object to the todo list..
		j.AddObjectToTodo( obj );

		// Move the object
		obj->a.x += r.x;
		obj->a.y += r.y;
		obj->b.x += r.x;
		obj->b.y += r.y;

		// Add the new position to the todo list
		j.AddObjectToTodo( obj );

		obj->Display();

		++ it;
  }
  
  j.CheckTodoList();
}


void CTinyCadDoc::AddNewJunction( CPoint q )
{
	// A junction is required, create it!
	CDrawJunction *junction = new CDrawJunction( this );
	junction->a = q;
	junction->b = q;

	m_drawing.insert( m_drawing.end(), junction );
	junction->Display();
}